/*
                       Sound Server Communications

	Functions:

	int SoundInit()
	int SoundChangeMode(char *arg)
	int SoundPlay(
		int code,
		double left_volume,
		double right_volume,
		int effects,
		int priority
	)
	int SoundManageEvents()
	void SoundShutdown()

	---

 */

#ifdef HAVE_YIFF
# include <Y2/Y.h>
# include <Y2/Ylib.h>
#endif	/* HAVE_YIFF */

#include "xsw.h"


/*
 *	Initializes the sound server, returns -1 on error.
 *
 *	Will return 0 and not initialize sound if option.sounds
 *	is XSW_SOUNDS_NONE.
 *
 *	sound.con_data must be NULL before calling this function!
 */
int SoundInit()
{
	int status;
	char stringa[1024];


	/* Initialize by which sound server type: */
	switch(sound.server_type)
	{
	  case SOUND_SERVER_TYPE_GENERIC:
	    break;

	  case SOUND_SERVER_TYPE_YIFF:
#ifdef Y_H
	    /* Connect to sound server. */
	    sound.con_data = (void *)YOpenConnection(
		sound.start_cmd,
		sound.con_arg
	    );
	    if(sound.con_data == NULL)
		return(-1);

	    /* Change mode. */
	    if(sound.mode_name[0] == '\0')
	        strcpy(sound.mode_name, "Default");
	    status = YChangeAudioModePreset(
		(YConnection *)sound.con_data,
	        sound.mode_name
	    );
	    if(status < 0)
	    {
		sprintf(stringa, "Cannot change sound server mode to `%s'.",
		    sound.mode_name
		);
		MesgAdd(stringa, xsw_color.standard_text);
	    }
#endif	/* Y_H */
	    break;

	  case SOUND_SERVER_TYPE_MSW:
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}


	/* Load sound schemes from file. */
	SSLoadFromFile(fname.sound_scheme);



	return(0);
}


/*
 *	Changes the mode to the one specified in arg.
 *	Sound server should already be initialized or else no operation
 *	is performed.
 */
int SoundChangeMode(char *arg)
{
	int status = 0;


        if(sound.con_data == NULL)
            return(-3);


        /* Change mode by which sound server type: */
        switch(sound.server_type)
        {
          case SOUND_SERVER_TYPE_GENERIC:
            break;
  
          case SOUND_SERVER_TYPE_YIFF:
#ifdef Y_H
	    if(arg == NULL)
		arg = "Default";

            status = YChangeAudioModePreset(
                (YConnection *)sound.con_data,
                arg		/* Mode name. */
            );
	    if(status < 0)
		status = -1;
	    else
		status = 0;
#endif	/* Y_H */
            break;
        
          case SOUND_SERVER_TYPE_MSW:
            break;

          case SOUND_SERVER_TYPE_MIKMOD:
            break;
        }


	return(status);
}



/*
 *	Plays a sound by given code number.
 */
int SoundPlay(
        int code,
        double left_volume,	/* 0.0 to 1.0 */
        double right_volume,	/* 0.0 to 1.0 */
        int effects,
        int priority		/* 0 or 1. */
)
{
	char *strptr;

#ifdef Y_H
	YID yid;
#endif	/* Y_H */


	if(sound.con_data == NULL)
	    return(-3);


	/* Get filename by sound code. */
	if(!SSIsAllocated(code))
	    return(-1);
	strptr = ss_item[code]->path;

        /* Play by which sound server type: */
        switch(sound.server_type)
        {
          case SOUND_SERVER_TYPE_GENERIC:
            break;  
 
          case SOUND_SERVER_TYPE_YIFF:
#ifdef Y_H
            yid = YStartPlaySoundObject(
		(YConnection *)sound.con_data,
		strptr,
		0,
		left_volume, right_volume,
		1		/* Repeats. */
	    );
#endif	/* Y_H */
            break;
  
          case SOUND_SERVER_TYPE_MSW:
            break;
   
          case SOUND_SERVER_TYPE_MIKMOD:
            break;
        }


	return(0);
}


/*
 *	Manages sound events if any, does not block.
 *
 *	Returns number of events handled.
 */
int SoundManageEvents()
{
	int events_handled = 0;

#ifdef Y_H
	YEvent yevent;
#endif	/* Y_H */


        switch(sound.server_type)
        {
          case SOUND_SERVER_TYPE_GENERIC:
            break;
 
          case SOUND_SERVER_TYPE_YIFF:
#ifdef Y_H
            if(sound.con_data != NULL)   
            {
		while(
		    YGetNextEvent(
                        (YConnection *)sound.con_data,
                        &yevent,
                        False			/* Do not block. */
                    ) > 0
		)
		{
		    events_handled++;

		    switch(yevent.type)
		    {
                      case YDisconnect:
                        YCloseConnection(
                            (YConnection *)sound.con_data,
                            False
                        );
                        sound.con_data = NULL;
			break;

                      case YSoundObjectKill:
                        break;

                      case YMixerChannel:
                        break;

                      case YSoundObjectPlay:
                        break;

                      case YShutdown:
                        YCloseConnection( 
                            (YConnection *)sound.con_data,
                            False
                        );
                        sound.con_data = NULL;
                        break;
		    }
		}
            }
#endif	/* Y_H */
            break;
  
          case SOUND_SERVER_TYPE_MSW:
            break;
 
          case SOUND_SERVER_TYPE_MIKMOD:
            break;
        }


	return(events_handled);
}


/*
 *	Shuts down the sound server.
 */
void SoundShutdown()
{
	/* Delete all sound schemes. */
	SSDeleteAll();


	/* Disconnect from sound server as needed. */
        switch(sound.server_type)
        {
          case SOUND_SERVER_TYPE_GENERIC:
            break;

          case SOUND_SERVER_TYPE_YIFF:
#ifdef Y_H
	    if(sound.con_data != NULL)
	    {
		YCloseConnection(
		    (YConnection *)sound.con_data,
		    False
		);
		sound.con_data = NULL;
	    }
#endif	/* Y_H */
            break;

          case SOUND_SERVER_TYPE_MSW:
            break;

          case SOUND_SERVER_TYPE_MIKMOD:
            break;
        }


	/* Set con_data to NULL just in case. */
	sound.con_data = NULL;


	return;
}
