/*
                Universe List Entry Management

	Functions:

	int UnivIsAllocated(int n)

	int UnivAdd(
		char *alias,
		char *url,
	        time_t last_connected,
	        char *comments,
	        int pos
	)        
	void UnivDelete(int n)
	void UnivDeleteAll()

	---

 */

#include "univlist.h"
#include "xsw.h"


#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


univ_entry_struct **univ_entry;
int total_univ_entries;



/*
 *	Checks if universe entry n is allocated.
 */
int UnivIsAllocated(int n)
{
        if((univ_entry == NULL) ||
           (n < 0) ||   
           (n >= total_univ_entries)
        )
        {
            return(0);
        }
        else
	{
	    return((univ_entry[n] == NULL) ? 0 : 1);
	}
}



/*
 *	Create a new universe entry and set given values to it.
 *
 *	Returns 0 on successs or -1 on error.
 */
int UnivAdd(
	char *alias,
	char *url,
	time_t last_connected,
	char *comments,
	int pos
)
{
        int i, new;
        univ_entry_struct *univ_ptr;


        /* Sanitize global variable total_univ_entries. */
        if(total_univ_entries < 0)
            total_univ_entries = 0;
         
        /* Increment global variable total_univ_entries. */
        total_univ_entries++;


        /* Check which entry style. */
        if(pos < 0)
        {
            /* Append entry. */
            
            /* Calculate new entry index. */
            new = total_univ_entries - 1;

            /* Allocate new entry. */
            univ_entry = (univ_entry_struct **)realloc(
		univ_entry,
                total_univ_entries * sizeof(univ_entry_struct *)
            );
            if(univ_entry == NULL)
            {
                total_univ_entries = 0;
                return(-1);
            }  
              
            univ_entry[new] = (univ_entry_struct *)calloc(1,
                sizeof(univ_entry_struct)
            );
            if(univ_entry[new] == NULL)
            {
                total_univ_entries--;
                return(-1);
            }
        }
        else
        {
            /* Insert. */

            /* Sanitize pos. */
            if(pos >= total_univ_entries)
                pos = total_univ_entries - 1;

            /* Allocate new entry. */
            univ_entry = (univ_entry_struct **)realloc(
		univ_entry,
                total_univ_entries * sizeof(univ_entry_struct *)
            );
            if(univ_entry == NULL)
            {
                total_univ_entries = 0;
                return(-1);
            }
                
            univ_ptr = (univ_entry_struct *)calloc(1,
                sizeof(univ_entry_struct)
            );
            if(univ_ptr == NULL)
            {
                total_univ_entries--;
                return(-1);
            }

            /* Shift entries. */
            for(i = total_univ_entries - 1; i > pos; i--)
            {
                univ_entry[i] = univ_entry[i - 1];
            }
            
            /* Adjust new entry number. */
            new = pos;

            /* Set new index to point to newly allocated pointer. */
            univ_entry[new] = univ_ptr;
        }


        /* Did we manage to allocate a new entry? */
        if(UnivIsAllocated(new))
	    univ_ptr = univ_entry[new];
	else
            return(-1);


	/* ******************************************************* */
        /* Set values. */

        if(alias != NULL)
        {
            univ_ptr->alias = (char *)calloc(1,
                MIN(strlen(alias) + 1,
                    UNIV_MAX_ALIAS_LEN + 1
		)
                * sizeof(char)
	    );
	    if(univ_ptr->alias != NULL)
	    {
                strncpy(univ_ptr->alias, alias,
                    MIN(strlen(alias), UNIV_MAX_ALIAS_LEN)
		);
	    }
        }
            
        if(url != NULL)
        {
            univ_ptr->url = (char *)calloc(1,
                MIN(strlen(url) + 1,
                    UNIV_MAX_URL_LEN + 1
                )
                * sizeof(char)
	    );
	    if(univ_ptr->url != NULL)
	    {
                strncpy(univ_ptr->url, url,
                    MIN(strlen(url), UNIV_MAX_URL_LEN)
		);
	    }
        }

        univ_ptr->last_connected = last_connected;

        if(comments != NULL)
        {
            univ_ptr->comments = (char *)calloc(1,
                MIN(strlen(comments) + 1,
                    UNIV_MAX_COMMENT_LEN + 1
                )
                * sizeof(char)
            );
	    if(univ_ptr->comments != NULL)
	    {
                strncpy(univ_ptr->comments, comments,
                    MIN(strlen(comments), UNIV_MAX_COMMENT_LEN)
		);
	    }
        }


        return(0);
}           



/*
 *	Delete universe entry n.
 */
void UnivDelete(int n)
{
        univ_entry_struct *entry_ptr;


        if(UnivIsAllocated(n))
            entry_ptr = univ_entry[n];
        else
            return;


        /* Free alias. */
#ifdef DEBUG_MEM_FREE
if(entry_ptr->alias != NULL)
    printf("Universe entry %i: Free'ed alias.\n", n);
#endif
        free(entry_ptr->alias);
        entry_ptr->alias = NULL;

        /* Free URL. */
#ifdef DEBUG_MEM_FREE
if(entry_ptr->url != NULL)
    printf("Universe entry %i: Free'ed url.\n", n);
#endif
        free(entry_ptr->url);
        entry_ptr->url = NULL;

        /* Reset last connected. */
        entry_ptr->last_connected = 0;

        /* Free comments. */
#ifdef DEBUG_MEM_FREE
if(entry_ptr->comments != NULL) 
    printf("Universe entry %i: Free'ed comments.\n", n);
#endif
        free(entry_ptr->comments);
        entry_ptr->comments = NULL;


#ifdef DEBUG_MEM_FREE
if(entry_ptr != NULL)
    printf("Universe entry %i: Free'ed.\n", n);
#endif
        free(entry_ptr);
        entry_ptr = NULL;


        return;
}


/*
 *	Procedure to delete all universe entries.
 */
void UnivDeleteAll()
{
        int i;


#ifdef DEBUG_MEM_FREE 
printf("Universe entry: Deleting %i...\n", total_univ_entries);
#endif

        for(i = 0; i < total_univ_entries; i++)
        {
            UnivDelete(i);
        }
  
#ifdef DEBUG_MEM_FREE
if(univ_entry != NULL)
    printf("Universe entry pointers: Free'ed.\n");
#endif
        free(univ_entry);
        univ_entry = NULL;

        total_univ_entries = 0;


        return;
}        

