/*
                       Object Database Management

	Functions:

        xsw_object_struct **DBAllocObjectPointers(
		xsw_object_struct **cur_obj_ptrs,
		long num_objects
        )
        xsw_object_struct *DBAllocObject()

        long DBCreateObject(
                int isref_num,
                int type,
                long owner,  
                double x,  
                double y,
                double z,
                double heading,
                double pitch,
                double bank
        )
        int DBCreateExplicitObject(
                long object_num, 
                int isref_num,
                int type,  
                long owner,
                double x, 
                double y,  
                double z,
                double heading,
                double pitch,
                double bank
        )
        long DBCreateObjectByOPM(  
                char *opmname,
                char *name,
                int type,
                double x,
                double y,
                double z,
                double heading,
                double pitch
        )

	void DBResetObject(xsw_object_struct *obj_ptr)
        void DBDeleteObject(xsw_object_struct *obj_ptr)
        void DBDeleteAllObjects()
        void DBRecycleObject(long object_num)

	void DBReclaim()

	---


 */

#include "swserv.h"


/*
 *	Reallocates given object pointer array.
 *
 *	If num_objects <= 0 then cur_obj_ptrs is free'ed if it is not
 *	NULL.
 */
xsw_object_struct **DBAllocObjectPointers(
        xsw_object_struct **cur_obj_ptrs,
        long num_objects   
)
{
        xsw_object_struct **obj_ptrs_rtn;
         
        
	if(num_objects > 0)
	{
	    /* Reallocate given pointer array. */
            obj_ptrs_rtn = (xsw_object_struct **)realloc(
                cur_obj_ptrs,
                num_objects * sizeof(xsw_object_struct *)
            );
            if(obj_ptrs_rtn == NULL)
            {
                fprintf(stderr,
 "DBAllocObjectPointers(): Error allocating %ld object pointers.\n",
                    num_objects
                );
            }
	}
	else
	{
	    /* Free given pointer array. */
	    free(cur_obj_ptrs);
	    cur_obj_ptrs = NULL;
	    obj_ptrs_rtn = NULL;
	}


        return(obj_ptrs_rtn);
}


/*
 *	Allocates an object structure, returns NULL on error.
 *
 *	If successful, the allocated object structure will be
 *	reset to default values by a call to DBResetObject().
 */
xsw_object_struct *DBAllocObject()
{
        xsw_object_struct *obj_ptr;

 
        /* Allocate a new object. */
        obj_ptr = (xsw_object_struct *)calloc(1,
            sizeof(xsw_object_struct)
        );
        if(obj_ptr == NULL)
        {
            fprintf(stderr,
                "DBAllocObject(): Memory allocation error.\n"
            );
            return(NULL);
        }


        /* Reset values on new object. */
        DBResetObject(obj_ptr);


        return(obj_ptr);
}



/*
 *      Creates a new object, returns its number or -1 on error.
 */
long DBCreateObject(
        int isref_num,
        int type,
        long owner,
        double x,
        double y,
        double z,
        double heading,  
        double pitch,
        double bank
)
{       
        long i, object_rtn;
        long prev_total_objects;
        xsw_object_struct *obj_ptr, **ptr;


        /* Type may not be garbage. */
        if(type <= XSW_OBJ_TYPE_GARBAGE)   
        {
            fprintf(stderr,
                "DBCreateObject: Error: Useless request to create garbage.\n"
            ); 
            return(-1);
        }       
        
        /* The isref_num must be valid. */
        if((isref_num < 0) || (isref_num >= ISREF_MAX))
        {
            fprintf(stderr,
 "DBCreateObject: Error: Invalid ISRef number %i.\n",
                isref_num
            );
            return(-1);
        }
        
        
	/* Go through XSW objects list. */        
        for(i = 0, ptr = xsw_object;
            i < total_objects;
            i++, ptr++
	)
        {
            if(*ptr == NULL)
		continue;

            /* Look for garbage objects. */
            if((*ptr)->type == XSW_OBJ_TYPE_GARBAGE)
                break;
        }
        if(i < total_objects)
        {
            /* object_count is the available object. */

            obj_ptr = xsw_object[i];

            DBResetObject(obj_ptr);
            obj_ptr->type = type;
            obj_ptr->owner = owner;
            obj_ptr->imageset = isref_num;
            obj_ptr->x = x;
            obj_ptr->y = y;
            obj_ptr->z = z;
            obj_ptr->heading = heading;
            obj_ptr->pitch = pitch;
            obj_ptr->bank = bank;
        
            obj_ptr->last_updated = cur_millitime;

            return(i);
        }

        /* No garbage objects, can we allocate more? */
        if((total_objects + OBJECT_ALLOCATE_AHEAD) > MAX_OBJECTS)
            return(-1);


        /* Allocate more objects in memory. */
        prev_total_objects = total_objects;
        total_objects += OBJECT_ALLOCATE_AHEAD;
                
        
        /* Allocate pointer array. */
        xsw_object = (xsw_object_struct **)DBAllocObjectPointers(
            xsw_object, total_objects
	);
        if(xsw_object == NULL)
        {
            total_objects = 0;
            return(-1);
        }
            
        
        /* Allocate each object. */
        for(i = prev_total_objects; i < total_objects; i++)
        {
            xsw_object[i] = (xsw_object_struct *)DBAllocObject();
            if(xsw_object[i] == NULL)
            {
                total_objects = i;
                return(-1);
            }
        }

        /* New object will be the first object allocated above. */
        object_rtn = prev_total_objects;   


        /* Set default values on the new object. */
        obj_ptr = xsw_object[object_rtn];

        obj_ptr->type = type;
        obj_ptr->owner = owner;
        obj_ptr->imageset = isref_num;
        obj_ptr->x = x;
        obj_ptr->y = y;
        obj_ptr->z = z;
        obj_ptr->heading = heading;
        obj_ptr->pitch = pitch;
        obj_ptr->bank = bank;

        obj_ptr->last_updated = cur_millitime;


        /* Return the number of the newly created object. */
        return(object_rtn);
}


/*
 *      Procedure to allocate an object explicitly, if the object
 *	at the specified index object_num exists, it will be
 *	recycled and then recreated.  If the object does not exist
 *	then it will be created.
 */
int DBCreateExplicitObject(
        long object_num,
        int isref_num,
        int type,
        long owner,
        double x,
        double y,
        double z,
        double heading,
        double pitch,
        double bank
)
{
	long i, object_rtn;
        xsw_object_struct *obj_ptr;


        /* Type may not be garbage or error. */
        if(type <= XSW_OBJ_TYPE_GARBAGE)
        {
            fprintf(stderr,
     "DBCreateExplicitObject(): Useless request to create garbage.\n"
            );
            return(-1);
        }
        /* Make sure object_num is a valid number. */
        else if((object_num < 0) || (object_num >= MAX_OBJECTS))
        {
            fprintf(stderr,
     "DBCreateExplicitObject(): Object number %ld out of range.\n",
                object_num
            );
            return(-1);
        }
        /* Make sure isref_num is a valid number. */
        else if((isref_num < 0) || (isref_num >= ISREF_MAX))
        {
            fprintf(stderr,
    "DBCreateExplicitObject(): ISRef number %i out of range.\n",
                isref_num
            );
            return(-1);
        }


        /* *********************************************************** */

        /* If object_num is not garbage, then recycle it. */
        if(!DBIsObjectGarbage(object_num))
            DBRecycleObject(object_num);


        /* Allocate memory if object_num is greater than total_objects. */
        if(object_num >= total_objects)
        {
            /* Can we allocate more? */
            if((object_num + OBJECT_ALLOCATE_AHEAD) >= MAX_OBJECTS)
            {
                fprintf(stderr,
 "DBCreateExplicitObject(): Maximum objects %i reached.\n",
		    MAX_OBJECTS
		);
                return(-3);
            }


            /* Reallocate objects pointer array. */ 
            xsw_object = (xsw_object_struct **)DBAllocObjectPointers(
                xsw_object,
                object_num + OBJECT_ALLOCATE_AHEAD + 1
            );
            if(xsw_object == NULL)
            { 
                total_objects = 0;
                return(-1);
            }

            /* Allocate each new object. */
            for(i = total_objects;
                i <= (object_num + OBJECT_ALLOCATE_AHEAD);
                i++
            )
            {
                xsw_object[i] = (xsw_object_struct *)DBAllocObject();
                if(xsw_object[i] == NULL)
                {
                    total_objects = i;
                    return(-1);
                }
            }

            /* Adjust global total. */
            i = total_objects;
            total_objects = object_num + OBJECT_ALLOCATE_AHEAD + 1;

	    /* Reset newly created objects. */
            while(i < total_objects)
            {
                DBResetObject(xsw_object[i]);
                i++;
            } 
        }


        /* Create explicit object. */
        object_rtn = object_num;

        /* Set default values on the new object. */
        DBResetObject(xsw_object[object_rtn]);

        obj_ptr = xsw_object[object_rtn];

        obj_ptr->type = type; 
        obj_ptr->owner = owner;
        obj_ptr->imageset = isref_num;
        obj_ptr->x = x;
        obj_ptr->y = y;
        obj_ptr->z = z;
        obj_ptr->heading = heading;
        obj_ptr->pitch = pitch;
        obj_ptr->bank = bank;

        obj_ptr->last_updated = cur_millitime;


        return(0);
}


/*
 *      Create an object by Object parameter macro referance.
 */
long DBCreateObjectByOPM(
        char *opmname,
        char *name,
        int type,
        double x,
        double y,
        double z,
        double heading,
        double pitch
)
{
        long object_num;  
        long opm_num;     


        /* Error checks. */
        if((opmname == NULL) ||
           (type <= XSW_OBJ_TYPE_GARBAGE)
        )
            return(-1);


        /* Get OPM num. */
        opm_num = OPMGetByName(opmname, type);
        if(OPMIsGarbage(opm_num))
        {
            fprintf(stderr,
                "DBCreateObjectByOPM(): %s: No such OPM.\n",
                opmname
            );
            return(-1);  
        }


        /* Create new object. */
        object_num = DBCreateObject(
            ISREF_DEFAULT,
            type,
            -1,         /* No owner. */
            x,
            y,
            z,
            heading,
            pitch,
            0.0000
        );
        if(DBIsObjectGarbage(object_num))
        {
            fprintf(stderr,
                "DBCreateObjectByOPM(): Cannot create object.\n"
            );
            return(-1);
        }


        /* Set OPM values from opm_num to Object object_num. */
        OPMModelObject(object_num, opm_num);
            
                
        /* Set values. */
        if(name != NULL)
            strncpy(xsw_object[object_num]->name, name, XSW_OBJ_NAME_MAX);
        else
            xsw_object[object_num]->name[0] = '\0';
	xsw_object[object_num]->name[XSW_OBJ_NAME_MAX - 1] = '\0';

        xsw_object[object_num]->x = x; 
        xsw_object[object_num]->y = y;
        xsw_object[object_num]->z = z;

        xsw_object[object_num]->heading = heading;
        xsw_object[object_num]->pitch = pitch;
        xsw_object[object_num]->bank = 0.0000;

        xsw_object[object_num]->animation.last_interval = cur_millitime;


        /* Return the number of the created object. */
        return(object_num);
}



/*
 *      Resets values of object given by pointer.
 */         
void DBResetObject(xsw_object_struct *obj_ptr)
{
        int i;


	/* obj_ptr is assumed valid. */
        if(obj_ptr == NULL)
            return; 


        /* ************************************************************** */

        /* Deallocate weapons. */
        for(i = 0; i < obj_ptr->total_weapons; i++)
        {
            if(obj_ptr->weapons[i] == NULL)
		continue;

#ifdef DEBUG_MEM_FREE
if(obj_ptr->weapons[i] != NULL)
    printf("Object weapon %i: Free'ed.\n", i);
#endif   
            free(obj_ptr->weapons[i]);
        }
#ifdef DEBUG_MEM_FREE
if(obj_ptr->weapons != NULL)
    printf("Object weapon pointers: Free'ed.\n");
#endif     
        free(obj_ptr->weapons);
        obj_ptr->weapons = NULL;

        obj_ptr->total_weapons = 0;
        obj_ptr->selected_weapon = -1;


        /* Deallocate tractored objects. */
#ifdef DEBUG_MEM_FREE
if(obj_ptr->tractored_object != NULL)
    printf("Object tractored objects: Free'ed.\n");
#endif
        free(obj_ptr->tractored_object);
        obj_ptr->tractored_object = NULL;
        obj_ptr->total_tractored_objects = 0;


        /* Deallocate score structure. */
#ifdef DEBUG_MEM_FREE
if(obj_ptr->score != NULL)
    printf("Object scores: Free'ed.\n");
#endif
        free(obj_ptr->score);
        obj_ptr->score = NULL;


        /* Deallocate economy structure. */
        if(obj_ptr->eco != NULL)
        {
            for(i = 0; i < obj_ptr->eco->total_products; i++)
            {
#ifdef DEBUG_MEM_FREE   
if(obj_ptr->eco->product[i] != NULL)
    printf("Object eco product %i: Free'ed.\n", i);
#endif
                free(obj_ptr->eco->product[i]); 
                obj_ptr->eco->product[i] = NULL;
            }
#ifdef DEBUG_MEM_FREE
if(obj_ptr->eco->product[i] != NULL)
    printf("Object eco product pointers: Free'ed.\n");
#endif
            free(obj_ptr->eco->product);
            obj_ptr->eco->product = NULL;
        }

#ifdef DEBUG_MEM_FREE
if(obj_ptr->eco != NULL)
    printf("Object eco: Free'ed.\n");
#endif     
        free(obj_ptr->eco);
        obj_ptr->eco = NULL;


        /* ************************************************************** */

        /* Reset values to that of hte garbage_object. */
	memcpy(
	    obj_ptr,		/* Destination. */
	    &garbage_object,	/* Source. */
	    sizeof(xsw_object_struct)
	);


        return;
}


/*
 *      Deletes object's allocated substructures and itself.
 */
void DBDeleteObject(xsw_object_struct *obj_ptr)
{
        if(obj_ptr == NULL)
	    return;


        /* Free all allocated substructures. */
        DBResetObject(obj_ptr);

        /* Free memory for givin object. */
        free(obj_ptr);
        obj_ptr = NULL;


        return;
}



/*
 *      Procedure to delete all objects referanced in the xsw_object
 *      pointer array.
 */     
void DBDeleteAllObjects()
{
        long i;


#ifdef DEBUG_MEM_FREE
printf("Objects: Deleting %i...\n", total_objects);
#endif

        for(i = 0; i < total_objects; i++)
        {
#ifdef DEBUG_MEM_FREE
if(xsw_object[i] != NULL)
    printf("Object %i: Free'ed.\n", i);
#endif
            DBDeleteObject(xsw_object[i]);
            xsw_object[i] = NULL;
        }

#ifdef DEBUG_MEM_FREE
if(xsw_object != NULL)
    printf("Object pointers: Free'ed.\n");
#endif
        free(xsw_object);
        xsw_object = NULL;

        total_objects = 0;


        return;
} 


/*
 *      Procedure to delete all allocated substructures on object
 *      and to set the object as garbage. 
 */
void DBRecycleObject(long object_num)   
{
	long i;
        char name1[256];
        char name2[256];

        int condescriptor;
        char stringa[1024 + XSW_OBJ_NAME_MAX];
        xsw_object_struct *obj_ptr, **ptr;


        /* Is object_num valid? */
        if(DBIsObjectGarbage(object_num))
            return;
        else
            obj_ptr = xsw_object[object_num];

        
        /* ********************************************************** */
        
        /* If object was a player, do some checks first. */
        if(obj_ptr->type == XSW_OBJ_TYPE_PLAYER)
        {
            /*   Change ownership of all objects owned by this player
             *   object to the global lost_found_owner object.
             */
            if(!DBIsObjectGarbage(sysparm.lost_found_owner))
            {
		/* Go through XSW objects list. */
                for(i = 0, ptr = xsw_object;
                    i < total_objects;
                    i++, ptr++
                )
                {
                    if(*ptr == NULL)
                        continue;
                    if((*ptr)->type <= XSW_OBJ_TYPE_GARBAGE)
                        continue;
                    if((*ptr)->owner != object_num)
                        continue;

		    /* Skip object itself. */
		    if(i == object_num)
			continue;

                    (*ptr)->owner = sysparm.lost_found_owner;
 
                    strncpy(name1,
			DBGetFormalNameStr(i), 256
		    );
                    name1[255] = '\0';
                    strncpy(name2,
			DBGetFormalNameStr(sysparm.lost_found_owner), 256
		    );
                    name2[255] = '\0';

                    sprintf(stringa,
               "Chowned %s to %s due to recycling of previous owner.",
                        name1, name2
                    );
                    if(sysparm.log_general)
                        LogAppendLineFormatted(fname.primary_log, stringa);
                }
            }
            /* If no valid lost and found owner, then set to owner 0. */
            else
            {
		/* Go through XSW objects list. */
                for(i = 0, ptr = xsw_object;
                    i < total_objects;
                    i++, ptr++
                )
                {
                    if(*ptr == NULL)
                        continue;
                    if((*ptr)->type <= XSW_OBJ_TYPE_GARBAGE)
                        continue;
                    if((*ptr)->owner != object_num)
                        continue;

                    /* Skip object itself. */
                    if(i == object_num)
                        continue;

                    (*ptr)->owner = 0;

                    sprintf(stringa,  
               "Chowned %s to #0 due to recycling of previous owner.",
                        DBGetFormalNameStr(i)
                    );
                    if(sysparm.log_general)
                        LogAppendLineFormatted(fname.primary_log, stringa);
                }
            }
                    
                        
            /* Close all connections logged in as object_num. */
	    while(1)
	    {
		condescriptor = ConGetByObject(object_num);
		if(condescriptor < 0)
		    break;

                NetCloseConnection(condescriptor);
            }
        }


        /* *********************************************************** */

        /* Reset and free all allocated substructures. */
        DBResetObject(obj_ptr);

        /* Setting object to type garbage to mark it as recycled. */  
        obj_ptr->type = XSW_OBJ_TYPE_GARBAGE;


        return;
}



/*         
 *      Frees outdated objects and reallocates object pointers.
 */        
void DBReclaim()
{
        long object_num;
        long top_object;



        /* Are there objects allocated? */
        if((total_objects <= 0) ||
           (xsw_object == NULL)
        )
            return;


        /* Get topmost object. */
        top_object = DBGetTopObjectNumber();
        if((top_object < 0) || (top_object >= total_objects))
            return;

        /* Do we need to reclaim memory? */
        if((total_objects - top_object - 1) <= OBJECT_ALLOCATE_AHEAD)
            return;


        /* Free all allocation from total_objects - 1 to top_object + 1. */
        for(object_num = total_objects - 1;
            object_num > top_object;
            object_num--
        )  
        {  
#ifdef DEBUG_MEM_FREE
if(xsw_object[object_num] != NULL)
    printf("Object %i: Free'ed (delete in reclaim).\n", object_num);
#endif
            DBDeleteObject(xsw_object[object_num]);
            xsw_object[object_num] = NULL;
        }


        /* Adjust global variable total_objects. */
        total_objects = top_object + 1;


        /* Reallocate object pointers. */
        xsw_object = (xsw_object_struct **)DBAllocObjectPointers(
            xsw_object, total_objects
        );
        if(xsw_object == NULL)
        {
            total_objects = 0;
            return;
        }


        return;
}
