/*
                      Math and Geometry Utility Functions

	Functions:

	double RADTODEG(double radians)
	double DEGTORAD(double degrees)
	double SANITIZERADIANS(double radians)

	double MuPolarRotX(double input_angle, double input_radius)
	double MuPolarRotY(double input_angle, double input_radius)

	void MuVectorAdd(
	        double v1_dir, double v1_mag,
        	double v2_dir, double v2_mag,
        	double *v_dir_rtn, double *v_mag_rtn
	)
 	double MuCoordinateDeltaVector(double dx, double dy)
 	double Mu3DDistance(double delta_x, double delta_y, double delta_z)

	int Mu3DInSector(long object_num, long sect_x, long sect_y, long sect_z)
	int Mu3DInSameSector(long object_num1, long object_num2)
        int Mu3DInSameSectorPtr(
                xsw_object_struct *obj_ptr1,
                xsw_object_struct *obj_ptr2
	)
 	int Mu3DInContact(long object_num1, long object_num2)
	int Mu3DInContactPtr(
		xsw_object_struct *obj_ptr1,
		xsw_object_struct *obj_ptr2
	)
        int Mu3DInVectorContact(
		long object_num1, long object_num2,
 		double heading, double heading_variance, double range
	)
 	int Mu3DInRange(long object_num1, long object_num2, double distance)
        int Mu3DInRangePtr(
		xsw_object_struct *obj_ptr1,
		xsw_object_struct *obj_ptr2,
		double distance
	)

 	double MuConvertSLUVelocityToWarp(double slu_velocity)
	double MuMaxRangeByVelocity(double velocity_max, long time)

	---

 	Math convience functions, all trigeometric functions use
 	the clock quadrant as referance and *not* the standard position
 	quadrant.    All angles must be in radians!   All distances
 	are in XSW 'real' units unless indicated otherwise.
 
 	MuPolarRotX() returns the x value for a point in polar
 	coordinates.   input_angle must be in radians.
 
 	MuPolarRotY() is the same as MuPolarRotX() except it returns
        the y value.
 
 	Mu3DDistance() returns the distance of two point's delta
        distances.
 
 	Mu3DInContact() returns 1 if XSW Objects object_num1 and
 	object_num2 are in contact, by calculating their distances
 	apart and their sizes.
 
 */

#include <stdio.h>
#include <db.h>
#include <sys/types.h>
#include <math.h>

#include "../include/reality.h"
#include "../include/objects.h"
#include "../include/mathutil.h"


extern int DBIsObjectGarbage(long object_num);


/*
 *	Converts radians to degrees.
 */
double RADTODEG(double radians)
{
        return(radians * 180 / PI);
}
        
/*
 *	Converts degrees to radians.
 */        
double DEGTORAD(double degrees)
{
        return(degrees * PI / 180);
}

/*
 *	Sanitizes radian value to: 0 <= radians < 2 * PI
 */
double SANITIZERADIANS(double radians)
{
#define RANGEINC(val,min,max)	(val > max) ? max : ((val < min) ? min : val)

	radians = RANGEINC(radians, -25.132741228, 25.132741228);

        while(radians >= 6.2831853)
            radians -= 6.2831853;
        
        while(radians < 0)
            radians += 6.2831853;

	return(radians);
}

double MuPolarRotX(double bearing_radians, double radius)
{
        return(
            radius *
            cos(SANITIZERADIANS((PI / 2) - bearing_radians))
        );
}

double MuPolarRotY(double bearing_radians, double radius)
{
        return(
	    radius *
	    sin(SANITIZERADIANS((PI / 2) - bearing_radians))
	);
}



void MuVectorAdd(
        double v1_dir, double v1_mag,
        double v2_dir, double v2_mag,
        double *v_dir_rtn, double *v_mag_rtn
)
{
        static double x1, y1, x2, y2, xr, yr;
        
 
        /* Returns must be valid. */
        if( (v_dir_rtn == NULL) ||
            (v_mag_rtn == NULL)
        )
            return;


        /* Magitudes must be positive. */
        if(v1_mag < 0)
            v1_mag *= -1;
        if(v2_mag < 0)
            v2_mag *= -1;

        /* Calculate right angle displacements. */
        x1 = sin(v1_dir) * v1_mag;
        y1 = cos(v1_dir) * v1_mag;
 
        x2 = sin(v2_dir) * v2_mag;
        y2 = cos(v2_dir) * v2_mag;
 
        
        /* Add vectors togeather. */
        xr = x1 + x2;
        yr = y1 + y2;


        *v_mag_rtn = hypot(xr, yr);
        *v_dir_rtn = MuCoordinateDeltaVector(xr, yr);


        return;
}



double MuCoordinateDeltaVector(double dx, double dy)
{
	return(SANITIZERADIANS((PI / 2) - atan2(dy, dx)));
}



double Mu3DDistance(double delta_x, double delta_y, double delta_z)
{
        /* Make sure all values are positive. */
        if(delta_x < 0)
            delta_x *= -1;
        if(delta_y < 0)
            delta_y *= -1;
        if(delta_z < 0)  
            delta_z *= -1;
         

        return(
	    hypot(hypot(delta_x, delta_y), delta_z)
	);
}



int Mu3DInSector(long object_num, long sect_x, long sect_y, long sect_z)
{
        static xsw_object_struct *obj_ptr;


        if(DBIsObjectGarbage(object_num))
            return(0);
	else
            obj_ptr = xsw_object[object_num];

	if((obj_ptr->sect_x == sect_x) &&
           (obj_ptr->sect_y == sect_y) &&
           (obj_ptr->sect_z == sect_z)
	)
	    return(1);
	else
	    return(0);
}


/*
 *	Checks if both objects are valid and in the same sector.
 */
int Mu3DInSameSector(long object_num1, long object_num2)
{
        static xsw_object_struct *obj1_ptr;
        static xsw_object_struct *obj2_ptr;


        /* object_num1 and object_num2 must be valid. */
        if(DBIsObjectGarbage(object_num1) ||
           DBIsObjectGarbage(object_num2)
        )
            return(0);


	obj1_ptr = xsw_object[object_num1];
        obj2_ptr = xsw_object[object_num2];


	return(
	    (obj1_ptr->sect_x == obj2_ptr->sect_x) &&
            (obj1_ptr->sect_y == obj2_ptr->sect_y) &&
            (obj1_ptr->sect_z == obj2_ptr->sect_z)
	);
}

int Mu3DInSameSectorPtr(
	xsw_object_struct *obj1_ptr,
	xsw_object_struct *obj2_ptr
)
{
        if((obj1_ptr == NULL) ||
           (obj2_ptr == NULL)
        )
            return(0);

	return(
            (obj1_ptr->sect_x == obj2_ptr->sect_x) &&
            (obj1_ptr->sect_y == obj2_ptr->sect_y) &&
            (obj1_ptr->sect_z == obj2_ptr->sect_z)
        );
}        


/*
 *	Checks if the two objects are in contact, takes size
 *	of each object into account.
 */
int Mu3DInContact(long object_num1, long object_num2)
{
        static double distance;
        static double size1, size2;
        static xsw_object_struct *obj1_ptr;
        static xsw_object_struct *obj2_ptr;


	/* Check if objects are valid and in the same sector. */
	if(!Mu3DInSameSector(object_num1, object_num2))
	    return(0);


        /* Get object pointers. */
        obj1_ptr = xsw_object[object_num1];
        obj2_ptr = xsw_object[object_num2];


        /* Both objects must have a size greater than 0. */
        size1 = obj1_ptr->size;		/* In Screen units. */
        size2 = obj2_ptr->size;
        size1 = (size1 < 0) ? 0 : size1;
        size2 = (size2 < 0) ? 0 : size2;


        /* Get the distance in Screen units. */
        distance = Mu3DDistance(
            obj2_ptr->x - obj1_ptr->x,
            obj2_ptr->y - obj1_ptr->y,
            obj2_ptr->z - obj1_ptr->z
        ) * 1000;			/* Convert to Screen units. */


        /* Is distance less than the two sizes? */
        if(distance < (size1 + size2))
            return(1);
        else
            return(0);
}

int Mu3DInContactPtr(
	xsw_object_struct *obj_ptr1, 
	xsw_object_struct *obj_ptr2
)
{
        static double distance;
        static double size1, size2;


	if((obj_ptr1 == NULL) ||
           (obj_ptr2 == NULL)
	)
	    return(0);

        if((obj_ptr1->sect_x != obj_ptr2->sect_x) ||
           (obj_ptr1->sect_y != obj_ptr2->sect_y) ||
           (obj_ptr1->sect_z != obj_ptr2->sect_z)
        )
            return(0);


        /* Both objects must have a size greater than 0. */
        size1 = obj_ptr1->size;		/* In Screen units. */
        size2 = obj_ptr2->size;
        size1 = (size1 < 0) ? 0 : size1;
        size2 = (size2 < 0) ? 0 : size2;


        /* Get the distance in Screen units. */
        distance = Mu3DDistance(
            obj_ptr2->x - obj_ptr1->x,
            obj_ptr2->y - obj_ptr1->y,
            obj_ptr2->z - obj_ptr1->z
        ) * 1000;                       /* Convert to Screen units. */


        /* Is distance less than the two sizes? */
        if(distance < (size1 + size2))
            return(1);
        else
            return(0);
}



int Mu3DInVectorContact(
        long object_num1,
        long object_num2,
        double heading,
        double heading_variance,
        double range
)
{
        static double distance;
        static double heading_result;
        static xsw_object_struct *obj1_ptr;
        static xsw_object_struct *obj2_ptr;
        

        /* Check if objects are valid and in the same sector. */
        if(!Mu3DInSameSector(object_num1, object_num2))
            return(0);


        /* range cannot be negative. */
        range = (range < 0) ? (range * -1) : range;


        /* Sanitize heading. */
	heading = SANITIZERADIANS(heading);
            
        /* heading_variance must be positive. */
        heading_variance = (heading_variance < 0) ?
            (heading_variance * -1) : heading_variance;

        /* Get object pointers. */
        obj1_ptr = xsw_object[object_num1];
        obj2_ptr = xsw_object[object_num2];

        
        /* ******************************************************** */
        /* Begin calculating. */
        
        /* Get the distance in pixel units. */
        distance = Mu3DDistance(
            obj2_ptr->x - obj1_ptr->x,
            obj2_ptr->y - obj1_ptr->y,
            obj2_ptr->z - obj1_ptr->z
        ) * 1000;               /* Convert to pixels. */
 
        
        /* If distance greater than range, return false. */
        if(distance > range)
            return(0); 
        
            
        /* At this point, it is within range. */
        
        /* Now see if it is within heading with heading_variance.   */
        heading_result = MuCoordinateDeltaVector(
            obj2_ptr->x - obj1_ptr->x,
            obj2_ptr->y - obj1_ptr->y
        );
        
        if( (heading_result >= (heading - heading_variance)) &&
            (heading_result <= (heading + heading_variance))
        )
        {
            return(1);
        }   
        else
        {
            return(0);
        }
}



/*
 *	Checks if the two objects are in the same sector and in range,
 *	both objects must be valid and distance is in XSW Real units.
 */
int Mu3DInRange(long object_num1, long object_num2, double distance)
{
        static double distance_apart;
        static double size1, size2;
        static xsw_object_struct *obj1_ptr;
        static xsw_object_struct *obj2_ptr;


	/* Check if objects are valid and in the same sector. */
        if(!Mu3DInSameSector(object_num1, object_num2))
            return(0);


        /* distance must be positive and convert to Screen units. */
        distance = (distance < 0) ?
            (distance * -1000) : (distance * 1000);
        
        
        /* Get object pointers. */
        obj1_ptr = xsw_object[object_num1];
        obj2_ptr = xsw_object[object_num2];


        /* Get sizes are in pixel units. */
        size1 = obj1_ptr->size;
        size2 = obj2_ptr->size;


        /* Get the distance in Screen units. */
        distance_apart = Mu3DDistance(
            obj2_ptr->x - obj1_ptr->x,
            obj2_ptr->y - obj1_ptr->y,
            obj2_ptr->z - obj1_ptr->z
        ) * 1000;
 

        /* See if distance apart is closer than distance. */
        if((distance_apart - (size1 + size2)) <= distance)
            return(1);
        else
            return(0);
}    


/*
 *      Checks if the two objects are in the same sector and in range,
 *      both objects must be valid and distance is in XSW Real units.
 */
int Mu3DInRangePtr(
	xsw_object_struct *obj_ptr1, 
        xsw_object_struct *obj_ptr2, 
        double distance 
)
{
        static double distance_apart;
        static double size1, size2;


	/* Check if objects are in the same sector. */
	if((obj_ptr1 == NULL) ||
           (obj_ptr2 == NULL)
	)
	    return(0);

	if((obj_ptr1->sect_x != obj_ptr2->sect_x) ||
           (obj_ptr1->sect_y != obj_ptr2->sect_y) ||
           (obj_ptr1->sect_z != obj_ptr2->sect_z)
	)
	    return(0);

        /* distance must be positive and convert to Screen units. */ 
        distance = (distance < 0) ?
            (distance * -1000) : (distance * 1000);


        /* Get sizes are in pixel units. */
        size1 = obj_ptr1->size;
        size2 = obj_ptr2->size;
  
  
        /* Get the distance in Screen units. */
        distance_apart = Mu3DDistance(
            obj_ptr2->x - obj_ptr1->x,
            obj_ptr2->y - obj_ptr1->y,
            obj_ptr2->z - obj_ptr1->z
        ) * 1000;


        /* See if distance apart is closer than distance. */
        if((distance_apart - (size1 + size2)) <= distance)
            return(1);
        else
            return(0);
}


/*
 *	Calculates the `theoretical' maximum range based on
 *	given maximum velocity v_max and time t.
 *
 *	If v_max is in XSW Real units, then the returned value will be
 *	in XSW Real units.  Same goes for XSW Screen units.
 *
 *	t must be in milliseconds.
 */
double MuMaxRangeByVelocity(double v_max, long t)
{
	return((v_max / (double)CYCLE_LAPSE_MS) * (double)t);
}
