/*
                    Reality Engine for XSW Objects

 	Functions:

        int REngFreqCmp(double f1, double f2)
	void REngMoveObject(long object_num, char allow_sect_change)

 	void REngHeading(long object_num)
 	void REngThrottle(long object_num)
 	void REngThrust(long object_num)
 	void REngLocation(long object_num)
	void REngVisibility(long object_num)
 	void REngShieldVisibility(long object_num)
 	int REngMortality(long object_num)
 	void REngAnimated(long object_num)
 	void REngWeaponsLock(long object_num)
	void REngTractorLock(object_num)
 	void REngIntercept(long object_num)
 	void REngPowerCore(long object_num)
 	void REngDamageControl(long object_num)
 	int REngDoHit(long wobj, long tobj)
 	int REngCollisionCheck(long object_num)
	int REngAI(long object_num)

 	int REngObject(long object_num)

	int REngInit()
	void REngManage()
	void REngShutdown()

	---

	Does all simulation (ie moving, changing, value adjusting)
	of XSW objects.

 */

#include "swserv.h"

#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#define MAX(a,b)        ((a) > (b) ? (a) : (b)) 


#define REngEventNone		0
#define REngEventMemChange	3


int REngFreqCmp(double f1, double f2);


/*
 *	Checks if the frequencies are the same up to 2 decimal
 *	accuracy.
 */
int REngFreqCmp(double f1, double f2)
{
	double df;


	/* Sanitize f1 and f2. */
	if(f1 < SWR_FREQ_MIN)
	    f1 = SWR_FREQ_MIN;
        if(f1 > SWR_FREQ_MAX)
            f1 = SWR_FREQ_MAX;

        if(f2 < SWR_FREQ_MIN)
            f2 = SWR_FREQ_MIN;
        if(f2 > SWR_FREQ_MAX)
            f2 = SWR_FREQ_MAX;

	/* Get absolute value difference. */
	df = f2 - f1;
	if(df < 0)
	    df *= -1;

	/* Compare up to 2 decimal places. */
	if(df < 0.01)
	    return(1);
	else
	    return(0);
}


/*
 *      Moves the object by calculating one quantom shift
 *      of its 3D coordinates.
 * 
 *      If allow_sect_change is true then if the object moves out
 *      if the coordinate bounds, it will change sectors and warp to
 *      the other side.
 */
void REngMoveObject(long object_num, char allow_sect_change)
{
        static char ch_sect_px, ch_sect_nx, ch_sect_py, ch_sect_ny;
        static int i;

        static double thrust_dir, thrust_mag, thrust_magc;
        static double velocity_mag, velocity_magc;
        static double x_vel, y_vel, z_vel;      /* Velocity delta. */
        static double x_velc, y_velc, z_velc;   /* Velocity delta /w time compensation. */
        static double x_thr, y_thr, z_thr;      /* Thrust. */
        static double x_thrc, y_thrc, z_thrc;   /* Thrust delta /w time compensation. */

	static double tmp_cos, tmp_sin;

        static xsw_object_struct *obj_ptr;
        static long owner_object;
        static long trac_object;
        static xsw_object_struct *owner_obj_ptr;
        static xsw_object_struct *trac_obj_ptr;


        /* Get object pointer (object_num assumed valid!). */
        obj_ptr = xsw_object[object_num];


        /* ************************************************************ */
        /* Stream weapons have location of their owner. */
        if((obj_ptr->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
           (obj_ptr->type == XSW_OBJ_TYPE_SPHEREWEAPON)
	)
        {
            owner_object = obj_ptr->owner;
            if(!DBIsObjectGarbage(owner_object))
            {
                owner_obj_ptr = xsw_object[owner_object];
                obj_ptr->x = owner_obj_ptr->x;
                obj_ptr->y = owner_obj_ptr->y;
                obj_ptr->z = owner_obj_ptr->z;

                obj_ptr->sect_x = owner_obj_ptr->sect_x;
                obj_ptr->sect_y = owner_obj_ptr->sect_y;
                obj_ptr->sect_z = owner_obj_ptr->sect_z;
            }
            return;
        }


        /* ***************************************************** */
        /* Standard object movement. */

        /* Sanitize velocity heading. */
        obj_ptr->velocity_heading = SANITIZERADIANS(obj_ptr->velocity_heading);


        /* Calculate movement from velocity (momentum).
         * Fetch velocity_mag, with compensation for lapsed time.
         */
        velocity_magc = obj_ptr->velocity * time_compensation;
        velocity_mag = obj_ptr->velocity;

        /* In quadrant 1? */
        if(obj_ptr->velocity_heading < 1.5707963)
        {
	   tmp_cos = cos(obj_ptr->velocity_heading);
	   tmp_sin = sin(obj_ptr->velocity_heading);

           x_velc = velocity_magc * tmp_sin;
           y_velc = velocity_magc * tmp_cos;
/*         z_velc = 0; */
           x_vel = velocity_mag * tmp_sin;
           y_vel = velocity_mag * tmp_cos;
/*         z_vel = 0; */
        }
        /* In Quadrant 2? */
        else if(obj_ptr->velocity_heading < 3.1415927 )
        {
           tmp_cos = cos(obj_ptr->velocity_heading - 1.5707963);
           tmp_sin = sin(obj_ptr->velocity_heading - 1.5707963) * -1;

           x_velc = velocity_magc * tmp_cos;
           y_velc = velocity_magc * tmp_sin;
/*         z_velc = 0; */
           x_vel = velocity_mag * tmp_cos;
           y_vel = velocity_mag * tmp_sin;
/*         z_vel = 0; */

        }
        /* In Quadrant 3? */
        else if(obj_ptr->velocity_heading < 4.712389 )
        {
           tmp_cos = cos(obj_ptr->velocity_heading - 3.1415927) * -1;
           tmp_sin = sin(obj_ptr->velocity_heading - 3.1415927) * -1; 

           x_velc = velocity_magc * tmp_sin;
           y_velc = velocity_magc * tmp_cos;
/*         z_velc = 0; */
           x_vel = velocity_mag * tmp_sin;
           y_vel = velocity_mag * tmp_cos;
/*         z_vel = 0; */
        }
        /* In Quadrant 4? */
        else
        {
           tmp_cos = cos(obj_ptr->velocity_heading - 4.712389) * -1;
           tmp_sin = sin(obj_ptr->velocity_heading - 4.712389);

           x_velc = velocity_magc * tmp_cos;
           y_velc = velocity_magc * tmp_sin;
/*         z_velc = 0; */
           x_vel = velocity_mag * tmp_cos;
           y_vel = velocity_mag * tmp_sin;
/*         z_vel = 0; */
        }


        /* Calculate movement from thrust. */
        thrust_dir = SANITIZERADIANS(
            obj_ptr->heading +
            obj_ptr->thrust_dir +
            PI
        );


        /* Get thrust_mag. */
	if(obj_ptr->power_max > 0)
            thrust_mag = obj_ptr->thrust * obj_ptr->power_purity *
		obj_ptr->power / obj_ptr->power_max;
	else
	    thrust_mag = obj_ptr->thrust * obj_ptr->power_purity;

        /* Sanitize thrust_mag. */
        if((thrust_mag + obj_ptr->velocity) >
            obj_ptr->velocity_max
        )
        {
           thrust_mag = obj_ptr->velocity_max - obj_ptr->velocity;
        }

        /* Adjust thrust with time_compensation. */
        thrust_magc = thrust_mag * time_compensation;


        /* In quadrant 1? */
        if(thrust_dir < 1.5707963)
        {
           x_thrc = thrust_magc * sin(thrust_dir);
           y_thrc = thrust_magc * cos(thrust_dir);
/*         z_thrc = 0; */
           x_thr = thrust_mag * sin(thrust_dir);
           y_thr = thrust_mag * cos(thrust_dir);
/*         z_thr = 0; */
        }
        /* In Quadrant 2? */
        else if(thrust_dir < 3.1415927 )
        {
           x_thrc = thrust_magc * cos(thrust_dir - 1.5707963);
           y_thrc = thrust_magc * sin(thrust_dir - 1.5707963) * -1;
/*         z_thrc = 0; */
           x_thr = thrust_mag * cos(thrust_dir - 1.5707963); 
           y_thr = thrust_mag * sin(thrust_dir - 1.5707963) * -1;
/*         z_thr = 0; */
        }
        /* In Quadrant 3? */
        else if(thrust_dir < 4.712389 )
        {
           x_thrc = thrust_magc * sin(thrust_dir - 3.1415927) * -1;
           y_thrc = thrust_magc * cos(thrust_dir - 3.1415927) * -1;
/*         z_thrc = 0; */
           x_thr = thrust_mag * sin(thrust_dir - 3.1415927) * -1;
           y_thr = thrust_mag * cos(thrust_dir - 3.1415927) * -1;
/*         z_thr = 0; */
        }
        /* In Quadrant 4? */
        else
        {
           x_thrc = thrust_magc * cos(thrust_dir - 4.712389) * -1;
           y_thrc = thrust_magc * sin(thrust_dir - 4.712389);
/*         z_thrc = 0; */
           x_thr = thrust_mag * cos(thrust_dir - 4.712389) * -1; 
           y_thr = thrust_mag * sin(thrust_dir - 4.712389);
/*         z_thr = 0; */
        }

        /* Move object. */
        obj_ptr->x += x_velc + x_thrc;
        obj_ptr->y += y_velc + y_thrc;
/*      obj_ptr->z += z_velc + z_thrc; */


        /* Do sector change as needed. */
        if(allow_sect_change)
        {
            ch_sect_px = 0;
            ch_sect_nx = 0;
            ch_sect_py = 0;
            ch_sect_ny = 0;

            if(obj_ptr->x < sector_legend.x_min)
            {
                obj_ptr->x += sector_legend.x_len;
                obj_ptr->sect_x -= 1;
                ch_sect_nx = 1;         /* Negative sect x change. */
	        NetSendObjectSect(-1, object_num);
            }
            else if(obj_ptr->x > sector_legend.x_max)
            {
                obj_ptr->x -= sector_legend.x_len;
                obj_ptr->sect_x += 1;
                ch_sect_px = 1;         /* Positive sect x change. */
                NetSendObjectSect(-1, object_num);
            }

            if(obj_ptr->y < sector_legend.y_min)
            {
                obj_ptr->y += sector_legend.y_len;
                obj_ptr->sect_y -= 1;
                ch_sect_ny = 1;         /* Negative sect y change. */
                NetSendObjectSect(-1, object_num);
            }
            else if(obj_ptr->y > sector_legend.y_max)
            {
                obj_ptr->y -= sector_legend.y_len;
                obj_ptr->sect_y += 1;
                ch_sect_py = 1;         /* Positive sect y change. */
                NetSendObjectSect(-1, object_num);
            }

/*
            if(obj_ptr->z < sector_legend.z_min)
            {
                obj_ptr->z += sector_legend.z_len;
                obj_ptr->sect_z -= 1;
                ch_sect_nz = 1;
                NetSendObjectSect(-1, object_num);
            }
            else if(obj_ptr->z > sector_legend.z_max)
            {
                obj_ptr->z -= sector_legend.z_len;
                obj_ptr->sect_z += 1;
                ch_sect_pz = 1;
                NetSendObjectSect(-1, object_num);
            }
*/

            /* Change sector for tractored objects. */ 
            if(ch_sect_nx || ch_sect_px ||
               ch_sect_ny || ch_sect_py
/*
               ch_sect_nz || ch_sect_pz
*/
            )   
            {
                for(i = 0; i < obj_ptr->total_tractored_objects; i++)
                {
                    trac_object = obj_ptr->tractored_object[i];
                    if(DBIsObjectGarbage(trac_object))
                        continue;
                    else
                        trac_obj_ptr = xsw_object[trac_object];
             
                    trac_obj_ptr->sect_x = obj_ptr->sect_x;
                    trac_obj_ptr->sect_y = obj_ptr->sect_y;
/*                  trac_obj_ptr->sect_z = obj_ptr->sect_z; */

		    /*   Move tractored object coordinates to exactly
		     *   where the source object is.
		     */
                    if(ch_sect_nx)
                        trac_obj_ptr->x = obj_ptr->x;
                    if(ch_sect_px)
                        trac_obj_ptr->x = obj_ptr->x;

                    if(ch_sect_ny)
                        trac_obj_ptr->y = obj_ptr->y;
                    if(ch_sect_py)
                        trac_obj_ptr->y = obj_ptr->y;


/*
                    if(ch_sect_nx)
                        trac_obj_ptr->x += sector_legend.x_len;
                    if(ch_sect_px)
                        trac_obj_ptr->x -= sector_legend.x_len;

                    if(ch_sect_ny)
                        trac_obj_ptr->y += sector_legend.y_len;
                    if(ch_sect_py)
                        trac_obj_ptr->y -= sector_legend.y_len;

                    if(ch_sect_nz)
                        trac_obj_ptr->z += sector_legend.z_len;
                    if(ch_sect_pz)
                        trac_obj_ptr->z -= sector_legend.z_len;
*/
		    NetSendObjectSect(-1, trac_object);
                }
            }
	}


        /*
         *   Record new velocity and velocity_heading on object.
         *   This value must not have the time compensation modifications
         *   to it.
         */
        obj_ptr->velocity_heading = MuCoordinateDeltaVector(
            x_vel + x_thr,
            y_vel + y_thr
        );

        obj_ptr->velocity = Mu3DDistance(
            x_vel + x_thr,
            y_vel + y_thr,
            0
        );


        return;
}



/*
 *      Updates the heding of the object.
 */
void REngHeading(long object_num)
{
	double x, y, z;
	double rel_heading, target_heading;
	long tobj_num;
	xsw_object_struct *obj_ptr, *tar_obj_ptr;


	/* Get object pointer. */
	obj_ptr = xsw_object[object_num];


	/* Skip static objects. */
        if(obj_ptr->type == XSW_OBJ_TYPE_STATIC)
           return;


	/*   Skip player objects that are not intercepting anything.
         *   When not intercepting, their heading is updated by the
	 *   client.
	 */
	if((obj_ptr->type == XSW_OBJ_TYPE_PLAYER) &&
	   (obj_ptr->intercepting_object < 0)
	)
           return;


	/* Is object intercepting something? */
	tobj_num = obj_ptr->intercepting_object;
	if(!DBIsObjectGarbage(tobj_num))
	{
	    /* Get rel_heading to the object we're intercepting. */
	    tar_obj_ptr = xsw_object[tobj_num];

	    if(Mu3DInSameSectorPtr(obj_ptr, tar_obj_ptr))
	    {
		/* In same sector. */

		/* Calculate over sight position. */
		x = sin(tar_obj_ptr->velocity_heading) *
		    tar_obj_ptr->velocity * time_compensation;
		y = cos(tar_obj_ptr->velocity_heading) *
                    tar_obj_ptr->velocity * time_compensation;

                target_heading = MuCoordinateDeltaVector(
                    (tar_obj_ptr->x + x) - obj_ptr->x,
                    (tar_obj_ptr->y + y) - obj_ptr->y
                );
	    }
	    else
	    {
		/* In different sectors. */

	        target_heading = MuCoordinateDeltaVector(
		    tar_obj_ptr->sect_x - obj_ptr->sect_x,
		    tar_obj_ptr->sect_y - obj_ptr->sect_y
	        );
	    }

	    rel_heading = SANITIZERADIANS(target_heading - obj_ptr->heading);

	    if((rel_heading < -3.1415927) || (rel_heading > 3.1415927))
	    {
	        /* rel_heading must be within the turnrate of object_num. */
	        if(rel_heading > (100 * obj_ptr->turnrate * time_compensation))
		    rel_heading = 100 * obj_ptr->turnrate * time_compensation;
	        else if(rel_heading < (-100 * obj_ptr->turnrate * time_compensation))
		    rel_heading = -100 * obj_ptr->turnrate * time_compensation;

		obj_ptr->heading -= rel_heading;
	    }
	    else
	    {
                /* rel_heading must be within the turnrate of object_num. */
                if(rel_heading > (100 * obj_ptr->turnrate * time_compensation))
                    rel_heading = 100 * obj_ptr->turnrate * time_compensation;
                else if(rel_heading < (-100 * obj_ptr->turnrate * time_compensation))
                    rel_heading = -100 * obj_ptr->turnrate * time_compensation;

                obj_ptr->heading += rel_heading;
	    }
	}

        /* Sanitize heading. */
	obj_ptr->heading = SANITIZERADIANS(obj_ptr->heading);


	return;
}



/*
 *	Nothing is done here, this function is just for client side.
 */
void REngThrottle(long object_num)
{
        xsw_object_struct *obj_ptr;


        /* Get object pointer. */
        obj_ptr = xsw_object[object_num];


        /* Skip static objects. */
        if(obj_ptr->type == XSW_OBJ_TYPE_STATIC)
           return;


        /* Skip player objects, their throttle is updated on client. */
        if(obj_ptr->type == XSW_OBJ_TYPE_PLAYER)
           return;


	return;
}



/*
 *	Thrust, antimatter consumption, and velocity drag 
 *	calculations are done here.
 */
void REngThrust(long object_num)
{
	double throttle;
	double thrust_power;
        xsw_object_struct *obj_ptr;


        /* Get object pointer. */
        obj_ptr = xsw_object[object_num];

	/*   Note:
	 *
	 *   We want to update all types of objects, since there is
	 *   no telling if an object should have power decrementation,
	 *   A civilization on a planet for instance, would use up
	 *   antimatter, so throttle applied would simulate antimatter
	 *   usage.
	 */

	/* Get throttle and thrust_power. */
        throttle = obj_ptr->throttle;
	if(throttle > 0)
	{
	    /* Get thrust power. */
            thrust_power = obj_ptr->thrust_power;

            /* Consume Antimatter. */
            obj_ptr->antimatter -= (throttle * thrust_power *
		time_compensation);
            if(obj_ptr->antimatter < 0)
                obj_ptr->antimatter = 0;

            /* Thrust output should be 0 if there is no antimatter left. */
            if(obj_ptr->antimatter <= 0)
                obj_ptr->thrust = 0;
	}


	/* Skip player objects from this point on. */
/*
        if(obj_ptr->type == XSW_OBJ_TYPE_PLAYER)
            return;
*/

	/* ******************************************************** */
	/* Thrust output and velocity drag. */

	/*   Thrust is calculated by taking the coefficent of the
	 *   throttle and multiplying that with the thrust_power.
	 */
	obj_ptr->thrust = MAX(
	    throttle * obj_ptr->thrust_power * time_compensation,
	    0
	);


	/* Make sure velocity is not greater than velocity_max. */
/*
	if(obj_ptr->velocity > obj_ptr->velocity_max)
	    obj_ptr->velocity = obj_ptr->velocity_max;
 */

        /* ******************************************************** */
        /* Drag. */   

        if(obj_ptr->velocity_max > 0)
            obj_ptr->velocity -= ((obj_ptr->velocity /
                obj_ptr->velocity_max) * obj_ptr->thrust_power *
                time_compensation
	    );
        else 
            obj_ptr->velocity = 0;


	/* If velocity is such a small number, then set it 0. */
	if(obj_ptr->velocity < 0.00003)
	    obj_ptr->velocity = 0;


	return;
}



/*
 *	Performs an update location procedure on the object.
 *	movement style depends on the type of the object and several
 *	other attributes.
 *
 *	REngMoveObject() will be called to do the actual moving.
 */
void REngLocation(long object_num)
{
	xsw_object_struct *obj_ptr;
        double distance;
        int i;
        xsw_object_struct **tar_obj_ptr;
        xsw_object_struct *nebula_obj_ptr = NULL;


        /* Get object pointer (assumed valid). */
        obj_ptr = xsw_object[object_num];


	/* Check through area type objects (if this object is not area). */
	if(obj_ptr->type != XSW_OBJ_TYPE_AREA)
	{
	    /* Go through XSW objects list. */
	    for(i = 0, tar_obj_ptr = xsw_object;
                i < total_objects;
                i++, tar_obj_ptr++
            )
            {
                if(*tar_obj_ptr == NULL)
                    continue;

                /* Check only objects of type XSW_OBJ_TYPE_AREA. */
                if((*tar_obj_ptr)->type != XSW_OBJ_TYPE_AREA)
                    continue;

                /* Skip same object. */
                if(*tar_obj_ptr == obj_ptr)
                    continue;


                /* What type of area? */
                switch((*tar_obj_ptr)->loc_type)
                {
                  case XSW_LOC_TYPE_NEBULA:
                    /* Not in same sector? */
                    if((obj_ptr->sect_x != (*tar_obj_ptr)->sect_x) ||
                       (obj_ptr->sect_y != (*tar_obj_ptr)->sect_y) ||
                       (obj_ptr->sect_z != (*tar_obj_ptr)->sect_z)
                    )
                        break;

                    distance = Mu3DDistance(
                        (*tar_obj_ptr)->x - obj_ptr->x,
                        (*tar_obj_ptr)->y - obj_ptr->y,
                        (*tar_obj_ptr)->z - obj_ptr->z
                    ) * 1000;
                    if(distance > (*tar_obj_ptr)->size)
                        break;

                    nebula_obj_ptr = *tar_obj_ptr;
                    break;

                  default:
                    break;
		}
            }

	    /* Is this object now in a nebula? */
	    if(nebula_obj_ptr != NULL)
            {
	        /* Object is now in a nebula. */

	        if(obj_ptr->loc_type != XSW_LOC_TYPE_NEBULA)
	        {
		    /* Previously not in nebula, so object just entered. */


	        }

	        obj_ptr->loc_type = XSW_LOC_TYPE_NEBULA;
	    }
	    else
	    {
	        /* Object is in normal space. */

		obj_ptr->loc_type = XSW_LOC_TYPE_SPACE;
	    }
	}



	/* Skip static and player objects. */
        if((obj_ptr->type == XSW_OBJ_TYPE_STATIC) ||
	   (obj_ptr->type == XSW_OBJ_TYPE_PLAYER) ||
           (obj_ptr->type == XSW_OBJ_TYPE_HOME) ||
           (obj_ptr->type == XSW_OBJ_TYPE_AREA) ||
           (obj_ptr->type == XSW_OBJ_TYPE_ANIMATED)
	)
            return;

	/* Move object. */
	REngMoveObject(object_num, 1);


	return;
}


/*
 *	Updates the object's current visiblity by checking its
 *	shield, cloak and location.
 */
void REngVisibility(long object_num)
{
	double v;
        xsw_object_struct *obj_ptr;


        /* Get object pointer, assumed valid. */
        obj_ptr = xsw_object[object_num];


        /* Are shields up? */
        if(obj_ptr->shield_state == SHIELD_STATE_UP)
        {
            if(obj_ptr->cur_visibility < 1)
            {
                obj_ptr->cur_visibility = MIN(
                    obj_ptr->cur_visibility + (0.01 * time_compensation),
                    1
                );
            }
        }            
        /* Is cloak up? */
        else if(obj_ptr->cloak_state == CLOAK_STATE_UP)
        {
            if(obj_ptr->power_max > 0)
                v = obj_ptr->visibility - (obj_ptr->cloak_strength *
                    obj_ptr->power_purity * obj_ptr->power /
                    obj_ptr->power_max);
            else
                v = obj_ptr->visibility;
                    
            if(obj_ptr->cur_visibility > v)
            {
                obj_ptr->cur_visibility = MAX(
                    obj_ptr->cur_visibility - (0.01 * time_compensation),
                    v
                );
            }
        }
        /* Adjust visibility to normal. */
        else
        {
            if(obj_ptr->cur_visibility > obj_ptr->visibility)
            {
                obj_ptr->cur_visibility = MAX(   
                    obj_ptr->cur_visibility - (0.01 * time_compensation),
                    obj_ptr->visibility
                );
            }
            else if(obj_ptr->cur_visibility < obj_ptr->visibility)
            {   
                obj_ptr->cur_visibility = MIN(
                    obj_ptr->cur_visibility + (0.01 * time_compensation),
                    obj_ptr->visibility
                );
            }
        }


	/* Sanitize. */
        if(obj_ptr->cur_visibility < 0)
            obj_ptr->cur_visibility = 0;
 	else if(obj_ptr->cur_visibility > 1)
	     obj_ptr->cur_visibility = 1;


	return;
}


/*
 *      Updates the object's shield visibility.
 */
void REngShieldVisibility(long object_num)
{
        static double x;
	static xsw_object_struct *obj_ptr;
        

	/* Get object pointer, assumed valid. */
	obj_ptr = xsw_object[object_num];


	/* Skip if shields not visable. */
	if(obj_ptr->shield_visibility <= 0)
	    return;


        /* Calculate deduction. */
        x = (double)((double)CYCLE_LAPSE_MS /
            (double)SHIELD_VISIBILITY_INTERVAL) * time_compensation; 

        /* Reduce shield visibility. */
        obj_ptr->shield_visibility -= x;

        /* Sanitize shield_visibility. */
        if(obj_ptr->shield_visibility < 0)
            obj_ptr->shield_visibility = 0;


        return;
}



/*
 *	Checks if the object is too old.
 */
int REngMortality(long object_num)
{
        xsw_object_struct *obj_ptr;


	/* Get object pointer (assumed valid). */
	obj_ptr = xsw_object[object_num];


        /*   Animation Cycle Check:
         *   If this is an animated object, check if all the frame
         *   cycles has cycled.   If they have, then recycle object.
         */
        if(obj_ptr->type == XSW_OBJ_TYPE_ANIMATED)
        {
	    if(obj_ptr->animation.cycle_times > 0)
	    {
                if(obj_ptr->animation.cycle_count >=
                    obj_ptr->animation.cycle_times
		)
                {
		    /* Recycle this object. */
                    DBRecycleObject(object_num);

		    /* Notify all connections about this recycle. */
		    NetSendRecycleObject(-1, object_num);

                    return(REngEventMemChange);
                }
	    }
        }


	/*   Aging Check:
	 *   Check if the object's age is older than lifespan,
	 *   If it is, then recycle it.
         */
        if(obj_ptr->lifespan > 0)
	{
            if((cur_millitime - obj_ptr->birth_time) >
                obj_ptr->lifespan
	    )
            {
		/* Recycle this object. */
                DBRecycleObject(object_num);

		/* Notify all connections about this recycle. */
	        NetSendRecycleObject(-1, object_num);

                return(REngEventMemChange);
            }
	}


	return(REngEventNone);
}



/*
 *	Updates animated object's animation frame position.
 */
void REngAnimated(long object_num)
{
	xsw_object_struct *obj_ptr;
 

        /* Get object pointer (assumed valid). */
        obj_ptr = xsw_object[object_num];


	/* Time to increment frame? */
        if((obj_ptr->animation.last_interval +
           obj_ptr->animation.interval) <= cur_millitime
	)
	{
	    obj_ptr->animation.current_frame++;

	    /* Looped? */
	    if(obj_ptr->animation.current_frame >=
		obj_ptr->animation.total_frames
	    )
	    {
		obj_ptr->animation.current_frame = 0;

		/* Increment cycled times only if there is a cycle limit. */
		if(obj_ptr->animation.cycle_times > 0)
		    obj_ptr->animation.cycle_count++;
	    }

	    obj_ptr->animation.last_interval = cur_millitime;
	}


	return;
}



/*
 *	Updates weapons lock on an object.
 */
void REngWeaponsLock(long object_num)
{
	long locked_object;
	xsw_object_struct *obj_ptr;


	/* object_num is assumed valid. */
	obj_ptr = xsw_object[object_num];


        /* Get locked_object. */
        locked_object = obj_ptr->locked_on;
	if(DBIsObjectGarbage(locked_object))
	{
	    /* Locked on object is now garbage, unlock. */
	    obj_ptr->locked_on = -1;

	    return;
	}


	/* Check distance. */
/* Need to take location of object into account? */
	if(Mu3DInRange(object_num, locked_object,
             obj_ptr->scanner_range *
	     DBGetObjectVisibility(locked_object)
	   )
	)
	{
	    /* Object still in range. */

	}
	else
	{
	    /* Object out of scanner range, unlock. */
	    obj_ptr->locked_on = -1;
	}

	return;
}



/*
 *	Updates an object's tractor beam lock on another object.
 */
void REngTractorLock(long object_num)
{
	int i;
	double delta_x, delta_y, delta_z;
	double theta1, theta2, theta3;
	double mag1, mag2, mag3;
	double distance;

	double old_velocity_max;

        long tractored_object;
        xsw_object_struct *obj_ptr, *tar_obj_ptr;  


        /* object_num is assumed valid. */
        obj_ptr = xsw_object[object_num];


	/* Skip if not tractoring anything. */
	if(obj_ptr->total_tractored_objects <= 0)
	    return;


        /* Unlock tractor beam if object is out of antimatter. */
        if(obj_ptr->antimatter <= 0)
        {
	    for(i = 0; i < obj_ptr->total_tractored_objects; i++)
	    {
		DBObjectUntractor(object_num, obj_ptr->tractored_object[i]);
		NetSendTractorBeamLock(-1, object_num, -1);
	    }
            return;
        }


	/* Move each tractored object. */
	for(i = 0; i < obj_ptr->total_tractored_objects; i++)
	{
	    tractored_object = obj_ptr->tractored_object[i];
	    if(tractored_object < 0)
		continue;

	    /* Check if tractored_object is not -1 but garbage. */
	    if(DBIsObjectGarbage(tractored_object))
	    {
		DBObjectUntractor(object_num, tractored_object);
		NetSendTractorBeamLock(-1, object_num, -1);
	        continue;
	    }
	    else
	    {
		tar_obj_ptr = xsw_object[tractored_object];
	    }

	    /* Check if target object is tractorable. */
	    if(!DBIsObjectTractorablePtr(obj_ptr, tar_obj_ptr))
	    {
		DBObjectUntractor(object_num, tractored_object);
	        NetSendTractorBeamLock(-1, object_num, -1);
                continue;
	    }

            /* Consume antimatter. */

/* For now consume a constant amount of antimatter based on object's size. */
obj_ptr->antimatter -= (
    (double)((double)tar_obj_ptr->size / 1000) * 0.05
    ) * time_compensation;

	    if(obj_ptr->antimatter < 0)
	        obj_ptr->antimatter = 0;


	    /* ********************************************************** */
	    /* Adjust object vector and move these types of objects. */
	    if(1
/*
	       (tar_obj_ptr->type == XSW_OBJ_TYPE_STATIC) ||
               (tar_obj_ptr->type == XSW_OBJ_TYPE_PLAYER) ||
	       (tar_obj_ptr->type == XSW_OBJ_TYPE_HOME) ||
               (tar_obj_ptr->type == XSW_OBJ_TYPE_AREA) ||
               (tar_obj_ptr->type == XSW_OBJ_TYPE_ANIMATED)
*/
	    )
	    {
		/* Get target object to source object delta distances. */
	        delta_x = obj_ptr->x - tar_obj_ptr->x;
	        delta_y = obj_ptr->y - tar_obj_ptr->y;
	        delta_z = obj_ptr->z - tar_obj_ptr->z;

	        distance = Mu3DDistance(delta_x, delta_y, delta_z);

		/*   Get target object current velocity and velocity
		 *   heading.
		 */
	        theta1 = tar_obj_ptr->velocity_heading;
	        mag1 = tar_obj_ptr->velocity;

		/* Target object to source object bearing. */
	        theta2 = MuCoordinateDeltaVector(delta_x, delta_y);

		/*   Calculate tractor beam strength by taking
		 *   the coefficient of the distance away the objects
		 *   are from each other.  The farther away they
		 *   are, the greater the tractor beam strength.
                 */
	        if(distance <= (double)((double)obj_ptr->size / 1000))
		    mag2 = 0;
	        else
	            mag2 = 0.009 * (double)distance /
			(double)MAX_TRACTOR_BEAM_LEN;

		/*   Add target object's current velocity vector with
		 *   the new velocity vector caused by the tractor beam
		 *   pull.
		 */
	        MuVectorAdd(
	            theta1, mag1,
	            theta2, mag2,
	            &theta3, &mag3
	        );

	        tar_obj_ptr->velocity_heading = theta3;
	        tar_obj_ptr->velocity = mag3;

	        /*   IMPORTANT:
		 *
	         *   We must temporarly increase the tractored object's 
	         *   maximum velocity before we move it and reset it
	         *   after moving it.
	         */
	        old_velocity_max = tar_obj_ptr->velocity_max;
	        if(tar_obj_ptr->velocity > tar_obj_ptr->velocity_max)
		    tar_obj_ptr->velocity_max = tar_obj_ptr->velocity;

	        /* Move the object. */
	        REngMoveObject(tractored_object, 0);

	        /* Restore original velocity_max. */
	        tar_obj_ptr->velocity_max = old_velocity_max;
	    }
	}



	return;
}


/*
 *	Only checks if object is in intercept range, does NOT
 *	adjust heading.
 */
void REngIntercept(long object_num)
{
        long intercepting_object;
        xsw_object_struct *obj_ptr;


        /* Get pointer to object (assumed valid). */
        obj_ptr = xsw_object[object_num];

	/* Get intercepting_object. */
	intercepting_object = obj_ptr->intercepting_object;
	if(DBIsObjectGarbage(intercepting_object))
	    return;

	/* Use MatchInterceptByNumber() to see if object is in range. */
	if(MatchInterceptByNumber(object_num, intercepting_object) < 0)
	    obj_ptr->intercepting_object = -1;

	return;
}



/*
 *	Regernerates an object's power core.
 */
void REngPowerCore(long object_num)
{
	xsw_object_struct *obj_ptr;


        /* Get pointer to object (assumed valid). */
        obj_ptr = xsw_object[object_num];

	/* Increase power in the object's power core. */
        if(obj_ptr->power < obj_ptr->power_max)
        {
            obj_ptr->power += (obj_ptr->core_efficency *
		obj_ptr->power_purity * time_compensation
	    );

            if(obj_ptr->power > obj_ptr->power_max)
                obj_ptr->power = obj_ptr->power_max;
        }

	return;
}



/*
 *	Repairs object if the object has damage control on.
 */
void REngDamageControl(long object_num)
{
	double power;
	double size;
	double hp_inc;
	xsw_object_struct *obj_ptr;


        /* Get pointer to object (assumed valid). */
        obj_ptr = xsw_object[object_num];


        /* Skip if damage control is not on. */
        if(obj_ptr->damage_control != DMGCTL_STATE_ON)
            return;

	/* Fetch and sanitize power. */
        power = obj_ptr->power;
	if(power < 0)
	    power = 0;

	/* Fetch and sanitize size. */
	size = obj_ptr->size;
	if(size < 1)
	    size = 1;

	/* Calculate hitpoints increment. */
	hp_inc = power / pow((size * 2), 3);
        if(hp_inc < 0.000001)		/* increment can't be too small. */
            hp_inc = 0.000001;
	hp_inc *= time_compensation;	/* Time correction. */


	/* Repair hitpoints. */
	obj_ptr->hp += hp_inc;
	if(obj_ptr->hp > obj_ptr->hp_max)
	    obj_ptr->hp = obj_ptr->hp_max;


	/* Consume antimatter based on how much repairs were made. */
/* Need better calculation. */
	obj_ptr->antimatter -= hp_inc;
	if(obj_ptr->antimatter < 0)
	    obj_ptr->antimatter = 0;


	return;
}



/*
 *	Performs a hit object procedure.
 *
 *	WARNING: This function makes no checks if wobj and tobj really
 *	hit each other!   It is up to REngCollisionCheck() to do all
 *	the contact checking.
 */
int REngDoHit(long wobj, long tobj)
{
	int i;
	double dx, theta;
	char stringa[256];
	char stringb[XSW_OBJ_NAME_MAX + 256];
	char stringc[XSW_OBJ_NAME_MAX + 256];

	long wobj_owner, new_object_num;
	connection_struct **con_ptr;

	double	damage_to_structure = 0,
		damage_to_shield = 0;

	int rtn_code = 0;


	/* Validate wobj and tobj. */
	if(DBIsObjectGarbage(wobj))
	    return(-1);
	if(DBIsObjectGarbage(tobj))
            return(-1);


	/* Weapon object must have power. */
	if(xsw_object[wobj]->power <= 0)
	    return(0);


        /* Log collision. */
        strncpy(stringb, DBGetFormalNameStr(tobj), XSW_OBJ_NAME_MAX + 256);
	stringb[XSW_OBJ_NAME_MAX + 256 - 1] = '\0';
	strncpy(stringc, DBGetFormalNameStr(wobj), XSW_OBJ_NAME_MAX + 256);
        stringc[XSW_OBJ_NAME_MAX + 256 - 1] = '\0';

        sprintf(stringa,
           "%s hit %s.",
           stringc, stringb
	);
        if(sysparm.log_events)
            LogAppendLineFormatted(fname.primary_log, stringa);


	/* *********************************************************** */

	/*   Make shields visiable on target object if the target
	 *   object has its shields up and the shield frequencies
	 *   on the target and weapon objects do NOT match.
	 */
	if((xsw_object[tobj]->shield_state == SHIELD_STATE_UP) &&
           !REngFreqCmp(xsw_object[tobj]->shield_frequency,
            xsw_object[wobj]->shield_frequency) &&
           (xsw_object[tobj]->power_max > 0)
	)
	{
	    xsw_object[tobj]->shield_visibility =
	        xsw_object[tobj]->power / xsw_object[tobj]->power_max;

	    /* Send shield visibility to all connections. */
	    NetSendShieldVis(-1, tobj);
	}

	/* Take away shield power and/or hitpoints from tobj. */
	if((sysparm.homes_destroyable) ? 1 :
           (xsw_object[tobj]->type != XSW_OBJ_TYPE_HOME)
	)
	{
	    /*   Target object has shields up and shield frequencies
             *   do not match?
             */
	    if((xsw_object[tobj]->shield_state == SHIELD_STATE_UP) &&
               !REngFreqCmp(xsw_object[tobj]->shield_frequency,
                xsw_object[wobj]->shield_frequency)
	    )
	    {
		/* Extra damage to shields due to power purity. */
		dx = MAX(
		    (xsw_object[wobj]->power
                        * (2 - xsw_object[tobj]->power_purity)) -
		        xsw_object[wobj]->power,
		    0
		);

                /*   Add damage to shields, increase damage due to
                 *   power_purity of target object.
                 */
                damage_to_shield += (xsw_object[wobj]->power + dx);

	        xsw_object[tobj]->power -= damage_to_shield;


		/* Shields went down, apply remaining damage to hull. */
	        if(xsw_object[tobj]->power < 0)
	        {
		    /* Remember values are negative. */

		    damage_to_shield += xsw_object[tobj]->power;

		    /*   Add damage to structure, do not include extra
		     *   damage caused by lack of power purity.
		     */
                    damage_to_structure -= (xsw_object[tobj]->power + dx);

		    xsw_object[tobj]->hp -= damage_to_structure;
	        }
	    }
	    /* Target object's shields were down or no shields at all. */
	    else
	    {
		damage_to_structure += xsw_object[wobj]->power;
	        xsw_object[tobj]->hp -= xsw_object[wobj]->power;
	    }
	}
	/* Sanitize target object's power (a must). */
	if(xsw_object[tobj]->power < 0)
	    xsw_object[tobj]->power = 0;


        /* *********************************************************** */
	/* Update score. */

	/* Damage givin by weapon object's owner. */
        wobj_owner = xsw_object[wobj]->owner;
	if(!DBIsObjectGarbage(wobj_owner))
	{
            if((sysparm.homes_destroyable) ? 1 :
               (xsw_object[tobj]->type != XSW_OBJ_TYPE_HOME)
            )
	    {
                if(!DBCreateObjectScores(wobj_owner))
	        {
		    /* Add damage_givin to weapon object owner. */
		    xsw_object[wobj_owner]->score->damage_givin +=
		        xsw_object[wobj]->power;

		    /* Add kills if target object hp's are less than 0. */
		    if(xsw_object[tobj]->hp <= 0)
		        xsw_object[wobj_owner]->score->kills++;
	        }
	    }
	}

	/* Damage recieved by target object. */
        if((sysparm.homes_destroyable) ? 1 :
           (xsw_object[tobj]->type != XSW_OBJ_TYPE_HOME)
        )
        {
            if(!DBCreateObjectScores(tobj))
            {
                /* Add damage_recieved to target object. */
                xsw_object[tobj]->score->damage_recieved +=
                    xsw_object[wobj]->power;
	    }

	    /* Log tobj's damage recieved. */
	    strncpy(stringb, DBGetFormalNameStr(tobj), XSW_OBJ_NAME_MAX + 256);
	    stringb[XSW_OBJ_NAME_MAX + 256 - 1] = '\0';
            strncpy(stringc, DBGetFormalNameStr(wobj_owner), XSW_OBJ_NAME_MAX + 256);
	    stringc[XSW_OBJ_NAME_MAX + 256 - 1] = '\0';

	    sprintf(stringa,
                "%s: Recieved %.2lf damage by weapon fired by %s",
	        stringb,
	        xsw_object[wobj]->power,
	        stringc
	    );
	    if(sysparm.log_events)
	        LogAppendLineFormatted(fname.primary_log, stringa);


	    /*   Send hit notify to target object and weapon object's
	     *   owner's connections.
	     */

	    /* Get bearing from target object to weapon object. */
            theta = MuCoordinateDeltaVector(
                xsw_object[wobj]->x - xsw_object[tobj]->x,
                xsw_object[wobj]->y - xsw_object[tobj]->y
            );
	    /* Bearing must be relative from target object's attitude. */
            theta = SANITIZERADIANS(theta - xsw_object[tobj]->heading);


	    /* Send hit notify to all connections in sector. */
	    for(i = 0, con_ptr = connection;
                i < total_connections;
                i++, con_ptr++
	    )
	    {
                if(*con_ptr == NULL)
                    continue;
                if((*con_ptr)->socket < 0)
                    continue;
                if((*con_ptr)->object_num < 0)
                    continue;

		/*   Is connection object and target object in same
		 *   sector?
		 */
		if(!Mu3DInSameSector((*con_ptr)->object_num, tobj))
		    continue;

                NetSendNotifyHit(
		    i,
		    wobj, tobj,
		    xsw_object[wobj]->power,	/* Total damage. */
		    theta,			/* Bearing, tobj rel. */
		    damage_to_structure,
		    damage_to_shield
		);
	    }
	}


        /* Create large explosion if tobj's hp are less than 0. */
	if(xsw_object[tobj]->hp <= 0)
	{
	    /* Big explosion. */
	    new_object_num = DBCreateObjectByOPM(
		"Explosion3",
		NULL,
		XSW_OBJ_TYPE_ANIMATED,
		xsw_object[tobj]->x,
		xsw_object[tobj]->y,
		xsw_object[tobj]->z,
		xsw_object[tobj]->heading,
		xsw_object[tobj]->pitch
	    );
	    if(!DBIsObjectGarbage(new_object_num))
	    {
		xsw_object[new_object_num]->sect_x = xsw_object[tobj]->sect_x;
                xsw_object[new_object_num]->sect_y = xsw_object[tobj]->sect_y;
                xsw_object[new_object_num]->sect_z = xsw_object[tobj]->sect_z;

		NetSendCreateObject(-1, new_object_num);
	    }

	    /* Recycle tobj. */
	    if(xsw_object[tobj]->permission.uid >
		ACCESS_UID_DESTROY_IMMUNE)
	    {
		/* Log destruction of tobj. */
		sprintf(stringa,
		    "%s: Destroyed due to hull structure colapse.",
		    DBGetFormalNameStr(tobj)
		);
	        if(sysparm.log_events)
		    LogAppendLineFormatted(fname.primary_log, stringa);

		/* Send notify about object being destroyed. */
		for(i = 0, con_ptr = connection;
                    i < total_connections;
                    i++, con_ptr++
                )
                {
                    if(*con_ptr == NULL)
                        continue;
                    if((*con_ptr)->socket < 0)
                        continue;
                    if((*con_ptr)->object_num < 0)
                        continue;

		    if((*con_ptr)->object_num == tobj)
		    {
                        NetSendSysMessage(
                            i,
                            CS_SYSMESG_CODE_WARNING,
                            "Your vessel has been destroyed."
                        );
		    }
		}


		/*   Recycle tobj and disconnect connection (if it
		 *   was a player).
		 *   IMPORTANT NOTE: tobj is now invalid!
		 */
                DBRecycleObject(tobj);

		/* Send recycle request to all connections for tobj. */
                NetSendRecycleObject(-1, tobj);
	    }
	    else
	    {
		/* Sanitize hp on tobj (since we didn't recycle it. */
                if(xsw_object[tobj]->hp < 0)
		    xsw_object[tobj]->hp = 0;
	    }

	    /* Target object destroyed, set return code to 1. */
	    rtn_code = 1;
	}
	else
	{
	    /* Create small explosion. */
	    new_object_num = -1;
	    switch(xsw_object[wobj]->type)
	    {
	      /* Stream weapon. */
	      case XSW_OBJ_TYPE_STREAMWEAPON:
		new_object_num = DBCreateObjectByOPM(
		    "Explosion1",
		    NULL,
		    XSW_OBJ_TYPE_ANIMATED,
		    xsw_object[tobj]->x,
		    xsw_object[tobj]->y,
		    xsw_object[tobj]->z,
		    xsw_object[tobj]->heading,
		    xsw_object[tobj]->pitch
		);
		break;

	      /* Sphere weapon. */
              case XSW_OBJ_TYPE_SPHEREWEAPON:  
                new_object_num = DBCreateObjectByOPM(
                    "Explosion1",
                    NULL,
                    XSW_OBJ_TYPE_ANIMATED,
                    xsw_object[tobj]->x,
                    xsw_object[tobj]->y,
                    xsw_object[tobj]->z,
                    xsw_object[tobj]->heading,
                    xsw_object[tobj]->pitch
                );
                break;

	      /* Other type of weapon. */
	      default:
                new_object_num = DBCreateObjectByOPM(
                    "Explosion1",
                    NULL,
                    XSW_OBJ_TYPE_ANIMATED,
                    xsw_object[wobj]->x,
                    xsw_object[wobj]->y,
                    xsw_object[wobj]->z,
                    xsw_object[wobj]->heading,
                    xsw_object[wobj]->pitch
                );
		break;
	    }
            if(DBIsObjectGarbage(new_object_num))
	    {
                sprintf(stringa,
        "REngDoHit(): Could not create small explosion object."
                );
                if(sysparm.log_errors)
                    LogAppendLineFormatted(fname.primary_log, stringa);
	    }
	    else
            {
                xsw_object[new_object_num]->sect_x = xsw_object[tobj]->sect_x;
                xsw_object[new_object_num]->sect_y = xsw_object[tobj]->sect_y;
                xsw_object[new_object_num]->sect_z = xsw_object[tobj]->sect_z;

		NetSendCreateObject(-1, new_object_num);                
            }
	}


	/* Recycle the weapon object if its not a stream weapon. */
        if(xsw_object[wobj]->type == XSW_OBJ_TYPE_WEAPON)
	{
            DBRecycleObject(wobj);
            NetSendRecycleObject(-1, wobj);

	    rtn_code = 1;
	}


	return(rtn_code);
}



/*
 *	Does collission check, returns number of objects that have
 *	been found to have collided.
 *
 *	If a collision was detected and the object pointers were
 *	reallocated, this function will return REngEventMemChange.
 *	Otherwise if there were no contacts or no reallocation of
 *	object memory, REngEventNone is returned.
 */
int REngCollisionCheck(long object_num)
{
	long object_count;
	double heading_variance, distance;
	int result;
	xsw_object_struct *obj_ptr, **target_obj_ptr;


	/* Get obj_ptr (assumed valid). */
	obj_ptr = xsw_object[object_num];


	/* Colission check: Standard weapon. */
        if(obj_ptr->type == XSW_OBJ_TYPE_WEAPON)
        {
	    /* Begin checking if object has come in contact. */

	    /* Go through XSW objects list. */
            for(object_count = 0, target_obj_ptr = xsw_object;
                object_count < total_objects;
                object_count++, target_obj_ptr++
            )
            {
		/* Skip following objects. */
                if(((*target_obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE) ||
		   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_SPHEREWEAPON) ||
		   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_AREA) ||
		   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_ANIMATED) ||
		/* Don't hit its owner. */
                    (obj_ptr->owner == object_count) ||
		/* Don't hit itself. */
		    (object_num == object_count)
		)
                    continue;

		/*   Does target's permission level make it immune to
		 *   being hit?
		 */
		if((*target_obj_ptr)->permission.uid <= ACCESS_UID_WEAPON_IMMUNE)
		    continue;


                /* Did object hit target object (come in contact)? */
                if(Mu3DInContact(object_num, object_count))
                {
		    REngDoHit(object_num, object_count);

                    return(REngEventMemChange);
                }
            }
        }
        /* Colission check: Stream weapon */
        else if(obj_ptr->type == XSW_OBJ_TYPE_STREAMWEAPON)
        {
	    /* Stream weapon object must have power. */
	    if(obj_ptr->power <= 0)
		return(REngEventNone);

	    /* Begin checking for contact. */

	    /* Go through XSW objects list. */
            for(object_count = 0, target_obj_ptr = xsw_object;
                object_count < total_objects;
                object_count++, target_obj_ptr++
	    )
            {
                /* Skip following objects. */
                if(((*target_obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE) ||
		   /* Don't hit itself. */
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_SPHEREWEAPON) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_AREA) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_ANIMATED) ||
                   /* Don't hit its owner. */
                   (obj_ptr->owner == object_count)
                )
                        continue;

                /*   Does target's permission level make it immune to
                 *   being hit?
                 */
                if((*target_obj_ptr)->permission.uid <= ACCESS_UID_WEAPON_IMMUNE)
                    continue;

		/* In same sector? */
	        if(!Mu3DInSameSectorPtr(obj_ptr, *target_obj_ptr))
		    continue;

                /*   Did it hit something? Get distance away
		 *   in screen units.
		 */
                distance = Mu3DDistance(
                    (*target_obj_ptr)->x - obj_ptr->x,
                    (*target_obj_ptr)->y - obj_ptr->y,
                    (*target_obj_ptr)->z - obj_ptr->z
                ) * 1000;
                if(distance > 0)
                { 
                    heading_variance = atan(
                        (*target_obj_ptr)->size / distance
                    );
                }
                else
                {
		    /*   No distance away, so partical hits partical
		     *   regardless of their bearings to each other.
		     */
                    heading_variance = (2 * PI);
                }
                result = Mu3DInVectorContact(
                        object_num,
                        object_count,
                        obj_ptr->heading,
                        heading_variance,
                        obj_ptr->size + (*target_obj_ptr)->size
                );
                if(result)
                {
                    REngDoHit(object_num, object_count);

                    /*   Instead of recycling the stream weapon object,
                     *   we set its power to 0.
                     *
                     *   Must referance this by index since REngDoHit()
                     *   might have realigned global object pointers!
                     */
                    xsw_object[object_num]->power = 0;

                    return(REngEventMemChange);
                }
	    }
	}
        /* Colission check: Sphere weapon */
        else if(obj_ptr->type == XSW_OBJ_TYPE_SPHEREWEAPON)
        {
            /* Stream weapon object must have power. */
            if(obj_ptr->power <= 0)
                return(REngEventNone);

            /* Begin checking for contact. */

	    /* Go through XSW objects list. */
            for(object_count = 0, target_obj_ptr = xsw_object;
                object_count < total_objects;
                object_count++, target_obj_ptr++
	    )
            {
                /* Skip following objects. */
                if(((*target_obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE) ||
                   /* Don't hit itself. */
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_STREAMWEAPON) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_SPHEREWEAPON) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_AREA) ||
                   ((*target_obj_ptr)->type == XSW_OBJ_TYPE_ANIMATED) ||
                   /* Don't hit its owner. */
                   (obj_ptr->owner == object_count)
                )
                        continue;
        
                /*   Does target's permission level make it immune to
                 *   being hit?
                 */
                if((*target_obj_ptr)->permission.uid <= ACCESS_UID_WEAPON_IMMUNE)
                    continue;

                /* In same sector? */
                if(!Mu3DInSameSectorPtr(obj_ptr, *target_obj_ptr))
                    continue;

                /* Did it hit something? */
                if(Mu3DInContact(object_num, object_count))
		{
                    REngDoHit(object_num, object_count);

                    /*   Instead of recycling the stream weapon object,
                     *   we set its power to 0.
                     *
                     *   Must referance this by index since REngDoHit()
                     *   might have realigned global object pointers!
                     */
                    xsw_object[object_num]->power = 0;

                    return(REngEventMemChange);
		}
	    }
	}

	return(REngEventNone);
}

/*
 *	Handles AI (Artificial Intelligence) for controlled objects.
 */
int REngAI(long object_num)
{
	int status = REngEventNone;
	int type, sel_wep, weapons_fired, iff_code;
	long i, closest_fobject, closest_uobject;
	xswo_ai_flags_t ai_flags;
	xsw_object_struct *obj_ptr, **ptr;
	xsw_weapons_struct *wep_ptr;
	double	scanner_range, distance,
		closest_fdistance, closest_udistance;
	double ovisibility;


        /* Get obj_ptr (assumed valid). */
        obj_ptr = xsw_object[object_num];


	/* Must be of type controlled. */
	if(obj_ptr->type != XSW_OBJ_TYPE_CONTROLLED)
	    return(REngEventNone);


	/* Reset closest objects. */
	closest_fobject = -1;
	closest_uobject = -1;

	/* Get this object's AI flags. */
	ai_flags = obj_ptr->ai_flags;

	/* Get this object's scanner range. */
	scanner_range = obj_ptr->scanner_range;
	closest_fdistance = scanner_range;
	closest_udistance = scanner_range;

	/* Go through objects list. */
	for(i = 0, ptr = xsw_object;
            i < total_objects;
            i++, ptr++
	)
	{
/* What about objects that need to follow objects across sectors? */

	    /* Skip ourselves. */
	    if(obj_ptr == *ptr)
		continue;

	    /* Checks if objects are valid and in the same sector. */
	    if(!Mu3DInSameSectorPtr(obj_ptr, *ptr))
		continue;

	    /* React to only these types of objects. */
	    type = (*ptr)->type;
	    if((type != XSW_OBJ_TYPE_CONTROLLED) &&
               (type != XSW_OBJ_TYPE_PLAYER) &&
               (type != XSW_OBJ_TYPE_WEAPON)
	    )
		continue;


	    /* Get other object's visibility. */
	    ovisibility = DBGetObjectVisibilityPtr(*ptr);
	    /* Calculate distance in XSW real units. */
	    distance = Mu3DDistance(
		(*ptr)->x - obj_ptr->x,
		(*ptr)->y - obj_ptr->y,
		(*ptr)->z - obj_ptr->z
	    );

	    /* Check IFF. */
	    iff_code = MatchIFFPtr(obj_ptr, *ptr);
	    if(iff_code == IFF_FRIENDLY)
	    {
		/* Friendly. */
		/* Range check */
	        if((distance > (scanner_range * ovisibility)) ||
	           (distance > closest_fdistance)
	        )
		    continue;

		closest_fdistance = distance;
		closest_fobject = i;
	    }
	    else
	    {
		/* Hostile or unknown. */
		/* Range check */
                if((distance > (scanner_range * ovisibility)) ||
                   (distance > closest_udistance)
                )
                    continue;

                closest_udistance = distance;
                closest_uobject = i;
	    }
	}


	/* ******************************************************* */
	/* Follow: Adjust intercepting object. */
	while((ai_flags & XSW_OBJ_AI_FIRE_ANY) ||
	      (ai_flags & XSW_OBJ_AI_FOLLOW_ANY)
	)
	{
	    /* Friendly has precidence over unknown or hostile. */
	    if(((ai_flags & XSW_OBJ_AI_FIRE_FRIEND) ||
                (ai_flags & XSW_OBJ_AI_FOLLOW_FRIEND)
               ) && (closest_fobject > -1)
	    )
	    {
		if(obj_ptr->intercepting_object != closest_fobject)
		{
		    obj_ptr->intercepting_object = closest_fobject;
		    obj_ptr->locked_on = closest_fobject;
		}
	    }
	    /* Hostile or unknown. */
	    else if(closest_uobject > -1)
	    {
                if(obj_ptr->intercepting_object != closest_uobject)
                {
                    obj_ptr->intercepting_object = closest_uobject;
                    obj_ptr->locked_on = closest_uobject;
                }   
            }
	    else
	    {
		if(obj_ptr->intercepting_object > -1)
		{
		    obj_ptr->intercepting_object = -1;
		    obj_ptr->locked_on = -1;
		}
	    }

            break;      /* Break out of this while() loop. */
	}
	/* ******************************************************* */
	/* Follow: Adjust throttle. */
	while(ai_flags & XSW_OBJ_AI_FOLLOW_ANY)
	{
	    /* Actions in this section need to corrilate with
	     * those in the above intercepting object check
	     * since `following' implies concensus in throttle
	     * and intercept.
	     */

            /* Friendly has precidence over unknown or hostile. */
	    if((ai_flags & XSW_OBJ_AI_FOLLOW_FRIEND) &&
               (closest_fobject > -1)
	    )
	    {
		/* Adjust throttle. */
		if(closest_fdistance > 0.6)
		{
		    obj_ptr->throttle += (0.02 * time_compensation);
		}
		else
		{
		    obj_ptr->throttle -= (0.02 * time_compensation);

		    /* Apply external dampers (brakes), this may
		     * be a bit stronger than it is equated on the
		     * client.
		     */
/*
                    obj_ptr->velocity -= (obj_ptr->thrust_power *
                        obj_ptr->power_purity * time_compensation
                    );
 */
                    obj_ptr->velocity -= (0.04 * time_compensation);
		    if(obj_ptr->velocity < 0)
			obj_ptr->velocity = 0;
		}

		/* Sanitize throttle value. */
		if(obj_ptr->throttle > 1)
		    obj_ptr->throttle = 1;
		else if(obj_ptr->throttle < 0)
		    obj_ptr->throttle = 0;
	    }
            /* Hostile or unknown. */
            else if(closest_uobject > -1)
            {
                /* Adjust throttle. */
                if(closest_udistance > 0.6)
		{
                    obj_ptr->throttle += (0.02 * time_compensation);
		}
                else
		{
                    obj_ptr->throttle -= (0.02 * time_compensation);

                    /* Apply external dampers (brakes), this may
                     * be a bit stronger than it is equated on the
                     * client.
                     */
/*
		    obj_ptr->velocity -= (obj_ptr->thrust_power *
                        obj_ptr->power_purity * time_compensation
                    );
 */
		    obj_ptr->velocity -= (0.04 * time_compensation);
                    if(obj_ptr->velocity < 0)
                        obj_ptr->velocity = 0;
		}

                /* Sanitize throttle value. */
                if(obj_ptr->throttle > 1)
                    obj_ptr->throttle = 1;
                else if(obj_ptr->throttle < 0)
                    obj_ptr->throttle = 0;
            }
            else
            {    
		obj_ptr->throttle = 0;
	    }

	    break;	/* Break out of this while() loop. */
	}
	/* ******************************************************* */
        /* Fire. */
        while(ai_flags & XSW_OBJ_AI_FIRE_ANY)
	{
	    /* Note: Procedure here is similar to the procedure
	     * in function NetHandleFireWeapon().
	     */

	    /* Get selected weapon pointer and check if its valid. */
	    sel_wep = obj_ptr->selected_weapon;
	    if((sel_wep < 0) || (sel_wep >= obj_ptr->total_weapons))
		break;

	    wep_ptr = obj_ptr->weapons[sel_wep];
	    if(wep_ptr == NULL)
		break;

	    /* Check if there is a target object in range.
	     * Note that this section is not from NetHandleFireWeapon()
	     */
	    if((ai_flags & XSW_OBJ_AI_FIRE_FRIEND) &&
               (closest_fobject > -1)
	    )
	    {
		if(closest_fdistance > ((double)wep_ptr->range / 1000))
		    break;
	    }
	    else if(closest_uobject > -1)
	    {
                if(closest_udistance > ((double)wep_ptr->range / 1000))
                    break;
            }
	    else
	    {
		break;
	    }


	    /* Too early to fire again? */
            if((wep_ptr->delay + wep_ptr->last_used) >= cur_millitime)
		break;

	    /* Cannot fire weapon while cloaked. */
	    if(obj_ptr->cloak_state == CLOAK_STATE_UP)
            {
		wep_ptr->last_used = cur_millitime;
		break;
	    }

            /* If weapon is expendable, see if object has any left. */
            if(wep_ptr->emission_type == WEPEMISSION_PROJECTILE)
            {
                if(wep_ptr->amount <= 0)
                    break;
	    }

            /* Does object have enough power to fire weapon? */
            if(obj_ptr->power < wep_ptr->create_power)
		break; 

            /* Create (fire) the weapon object. */
            /* Stream weapons. */
            if(wep_ptr->emission_type == WEPEMISSION_STREAM)
            {
                weapons_fired = WepCreate(
		    wep_ptr->ocs_code,
		    object_num,		/* Owner. */
		    wep_ptr->emission_type,
		    obj_ptr->x,
		    obj_ptr->y,
		    obj_ptr->heading,
		    wep_ptr->power,
		    wep_ptr->range,
		    SWR_FREQ_MIN	/* Minimal frequency for now. */
		);
            }
            /* Projectile and pulse weapons. */
            else
            {
		weapons_fired = WepCreate(
		    wep_ptr->ocs_code,
		    object_num,		/* Owner. */
		    wep_ptr->emission_type,
		    obj_ptr->x,
                    obj_ptr->y,
                    obj_ptr->heading,
                    wep_ptr->power,
                    wep_ptr->range,
		    SWR_FREQ_MIN	/* Minimal frequency for now. */
		);
	    }

            /*   Reget objects array pointer, creating objects
             *   changes the global xsw_object pointer array!!!
             */
            obj_ptr = xsw_object[object_num];

	    /* Update status to report memory change. */
	    status = REngEventMemChange;

            /* No objects created? */
            if(weapons_fired <= 0)
                break;


            /* Mark the last time it was fired. */
            wep_ptr->last_used = cur_millitime;


            /*   Consume power on object for the power that was
             *   needed to create this weapon.
             */
            switch(wep_ptr->emission_type)
            {
              case WEPEMISSION_PROJECTILE:
                obj_ptr->power -= (wep_ptr->create_power * weapons_fired);
                /* Decrease stock for projectiles. */
                wep_ptr->amount -= weapons_fired;
                if(wep_ptr->amount < 0)
                    wep_ptr->amount = 0;
		break;

	      case WEPEMISSION_PULSE:
                obj_ptr->power -= (wep_ptr->create_power * weapons_fired);
                break;
            
              /* Default to stream weapon. */
              default:
                obj_ptr->power -= (wep_ptr->create_power);
 	        break;
	    }
            /* Sanitize power on object. */
            if(obj_ptr->power < 0)
                obj_ptr->power = 0;

            /* Set next object values update to now so it gets updated now. */
/*
            next.object_values = cur_millitime;
            next.weapon_values = cur_millitime;
 */

	    break;	/* Break out of this while() loop. */
	}


	return(status);
}


/*
 *	Procedure to completely update the object.
 */
int REngObject(long object_num)
{
	int status;

	/* object_num is assumed valid and non-garbage. */


	REngVisibility(object_num);
	REngShieldVisibility(object_num);

	status = REngMortality(object_num);
	if(status == REngEventMemChange)
	{
	    /* Implies object was recycled. */
	    return(REngEventMemChange);
	}

        REngAnimated(object_num);

	REngPowerCore(object_num);
        REngDamageControl(object_num);

        REngWeaponsLock(object_num);
	REngTractorLock(object_num);
	REngIntercept(object_num);

	REngThrottle(object_num);
        REngThrust(object_num);

	REngHeading(object_num);
        REngLocation(object_num);

	status = REngCollisionCheck(object_num);
	if(status == REngEventMemChange)
	{
	    /* Implies object was recycled. */
            return(REngEventMemChange);
	}

	status = REngAI(object_num);
	if(status == REngEventMemChange)
        {
            return(REngEventMemChange);
        }


	return(REngEventNone);
}


/* ******************************************************************** */

/*
 *      Initializes reality engine.
 */
int REngInit()
{
        return(0);
}


/*
 *      Manages all reality engine responsibilities.
 */
void REngManage()
{
	int status;
	long object_num;
	xsw_object_struct **obj_ptr;


	/* Go through XSW objects list. */
        for(object_num = 0, obj_ptr = xsw_object;
            object_num < total_objects;
            object_num++, obj_ptr++
	)
	{
	    if(*obj_ptr == NULL)
		continue;
	    if((*obj_ptr)->type <= XSW_OBJ_TYPE_GARBAGE)
		continue;

	    status = REngObject(object_num);

	    /* Reget global objects pointer on memory change. */
	    if(status == REngEventMemChange)
		obj_ptr = &(xsw_object[object_num]);
	}

	return;
}


/*
 *      Shuts down reality engine.
 */
void REngShutdown() 
{
        return;
}
