/*
                         Widget: List Window

        Functions:

        int ListGetItemNumByDataPointer(
                list_window_struct *lw,
                void *data_ptr
        )

	int ListIsItemAllocated(
	        list_window_struct *lw,
	        int n
	)

        int ListAddItem(
                list_window_struct *lw,
                int type,
                char *name,
                image_t *image,
                int pos,
                void *data_ptr
        )
        void ListDeleteItem(list_window_struct *lw, int pos)
        void ListDeleteAllItems(list_window_struct *lw)


        int ListWinInit(
                list_window_struct *lw,
                win_t parent,
                int x, int y,
                unsigned int width, unsigned int height
        )
        int ListWinResize(list_window_struct *lw)
        int ListWinDraw(list_window_struct *lw)
        int ListWinManage(
                list_window_struct *lw,
                event_t *event
        )
	void ListWinMapCVPrompt(list_window_struct *lw, int mode)
	void ListWinUnmapCVPrompt(list_window_struct *lw)
        void ListWinMap(list_window_struct *lw);
        void ListWinUnmap(list_window_struct *lw);
        void ListWinDestroy(list_window_struct *lw)

	void ListWinRepeatRecordSet(
		list_window_struct *lw,  
		long dsec, int op_code
	)
	void ListWinRepeatRecordClear()
	int ListWinManageRepeat(event_t *event)

        ---


 */

#include "../include/string.h"

#include "../include/widget.h"



#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define MAX(a,b)        (((a) > (b)) ? (a) : (b))


/*
 *	Size constants (in pixels):
 */
#define LW_DEF_CHAR_WIDTH	7
#define LW_DEF_CHAR_HEIGHT	14

#define LW_DEF_ROW_HEIGHT	32

#define LW_PROMPT_HEIGHT	30


/*
 *	Maximum length of list item's name.
 */
#define LW_ITEM_NAME_MAX	256

/*
 *	Default name to give to an item who's name is NULL.
 */
#define LW_DEF_ITEM_NAME	"Untitled"

         
/*
 *      Drag tolorance in pixels.
 */
#ifndef LW_DRAG_TOLORANCE
# define LW_DRAG_TOLORANCE          6
#endif


/*
 *	List window repeat record:
 */
#define LW_OP_CODE_NONE         0
#define LW_OP_CODE_SCROLL_UP    1	/* One unit. */
#define LW_OP_CODE_SCROLL_DOWN  2	/* One unit. */

typedef struct {

	list_window_struct *lw;
	long next_repeat;		/* Next repeat in ms. */
	int op_code;			/* One of LW_OP_CODE_*. */

} lw_repeat_record_struct;
lw_repeat_record_struct lw_repeat_record[1];



/*
 *      Drag records.
 */
static int drag_start_entry_pos;
static win_t drag_start_w;     /* Which window did drag start on. */
static int drag_start_x, drag_start_y; 
static bool_t drag_active;
         
static bool_t button1_state;
static bool_t button2_state;
static bool_t button3_state;

           

/*
 *      Returns the index number of a list item on the list window
 *	widget lw that has a matching data pointer data_ptr.
 *
 *      Returns -1 if no match could be made.
 */
int ListGetItemNumByDataPointer(
        list_window_struct *lw,
        void *data_ptr
)
{
        int i;
	list_window_entry_struct **ptr;


        if((lw == NULL) ||
           (data_ptr == NULL)
        )
            return(-1);

  
        /* Scan through list. */
        for(i = 0, ptr = lw->entry;
            i < lw->total_entries;
            i++, ptr++
	)
        {
            if(*ptr == NULL)
		continue;

	    /* Data pointers match? */
            if((*ptr)->data_ptr == data_ptr)
		return(i);
        }


	return(-1);
}


/*
 *	Checks if item n is allocated on lw.
 */
int ListIsItemAllocated(
        list_window_struct *lw,
        int n
)
{
	if(lw == NULL)
	    return(0);
	else if((lw->entry == NULL) ||
                (n < 0) ||
                (n >= lw->total_entries)
	)
	    return(0);
	else if(lw->entry[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *      Allocates a mew list item on list window lw.
 *
 *      Returns -1 on error, 0 on success.
 */
int ListAddItem(
        list_window_struct *lw,
        int type,
        char *name,
        image_t *image,
        int pos,
        void *data_ptr
)
{
        int i, n;
        list_window_entry_struct *new_entry;


        if(lw == NULL)
            return(-1);

        /* If name is NULL, use default name. */
        if(name == NULL)
            name = LW_DEF_ITEM_NAME;

	/* Allocate a new entry. */
	new_entry = (list_window_entry_struct *)calloc(
	    1,
	    sizeof(list_window_entry_struct)
        );
	if(new_entry == NULL)
	    return(-1);


        /* Increment total entries on list widget. */
        if(lw->total_entries < 0)
            lw->total_entries = 0;
        lw->total_entries++;

        /* Allocate more entry pointers. */
        lw->entry = (list_window_entry_struct **)realloc(
            lw->entry,
            lw->total_entries * sizeof(list_window_entry_struct *)
        );
        if(lw->entry == NULL)
        {
	    free(new_entry);
            lw->total_entries = 0;

            return(-1); 
        }


        /* Append or insert? */
        if(pos < 0)
        {
            /* Append. */

	    /* Set new position n to the last position. */
            n = (int)lw->total_entries - 1;
        }
        else
        {
            /* Insert. */

            /* Set new position n. */
            n = pos;

            /* Sanitize n. */
            if(n >= lw->total_entries)
                n = lw->total_entries - 1;
            if(n < 0)
                n = 0;


            /* Shift entry pointers. */
            for(i = lw->total_entries - 1; i > n; i--)
                lw->entry[i] = lw->entry[i - 1];
        }

	/* Set new entry pointer to point to the just allocated struct. */
	lw->entry[n] = new_entry;


	/* Set new values. */
        new_entry->type = type;
	new_entry->name = StringCopyAlloc(name);
        new_entry->image = image;
        new_entry->data_ptr = data_ptr;


        return(0);
}



/*
 *      Deletes entry pos on list and reallocates the entry pointers.
 */
void ListDeleteItem(list_window_struct *lw, int pos)
{
        int i;


        if(lw == NULL)
            return;

        if((pos < 0) || (pos >= lw->total_entries))
            return;

	/* Delete entry at pos. */
        if(lw->entry[pos] != NULL)
        {
            /* Free entry and its allocated substructures. */
            free(lw->entry[pos]->name);

            free(lw->entry[pos]);
            lw->entry[pos] = NULL;
        }

        /* Shift entries. */
        for(i = pos; i < (lw->total_entries - 1); i++)
            lw->entry[i] = lw->entry[i + 1];

        /* Reallocate entry pointers. */
        lw->total_entries--;
        if(lw->total_entries > 0)
        {
            lw->entry = (list_window_entry_struct **)realloc(
		lw->entry,
                lw->total_entries * sizeof(list_window_entry_struct *)
            );
            if(lw->entry == NULL)
            {
                lw->total_entries = 0;
                return;
            }
        }
	else
	{
	    free(lw->entry);
	    lw->entry = NULL;

	    lw->total_entries = 0;
	}


        /* Sanitize selected entry position. */
        if(lw->entry_pos >= lw->total_entries)
            lw->entry_pos = lw->total_entries - 1;
        if(lw->entry_pos < 0)
            lw->entry_pos = -1; /* Not selected. */


        return;
}


/*
 *	Deletes all entries on list window widget lw.
 */
void ListDeleteAllItems(list_window_struct *lw)
{
        int i;
	list_window_entry_struct **ptr;


        if(lw == NULL)
            return;

        for(i = 0, ptr = lw->entry;
            i < lw->total_entries;
            i++, ptr++
	)
        {
            if(*ptr == NULL)
		continue;

            free((*ptr)->name);
            free(*ptr);
        }

        free(lw->entry);
        lw->entry = NULL;

        lw->total_entries = 0;
        lw->entry_pos = -1;


        return;
}



/*
 *      Initialize a list widget.
 */     
int ListWinInit(
        list_window_struct *lw,
        win_t parent,
        int x, int y,
        unsigned int width, unsigned int height
)
{
        if((lw == NULL) ||
	   (parent == 0)
        )
            return(-1);


        /* Sanitize width and height. */
        if(width < 2)
            width = 2;
        if(height < 2)
            height = 2;


	/* Reset values. */
        lw->map_state = 0;
        lw->visibility_state = VisibilityFullyObscured;
        lw->is_in_focus = 0;
	lw->x = x;
	lw->y = y;
        lw->width = width;
        lw->height = height;
	lw->font = OSWQueryCurrentFont();
	lw->next = NULL;
	lw->prev = NULL;

	lw->allow_drag = True;

        lw->char_width = LW_DEF_CHAR_WIDTH;
        lw->row_height = LW_DEF_ROW_HEIGHT;

        lw->total_entries = 0;
        lw->entry = NULL;
        lw->entry_pos = -1;

        lw->left_margin = 12;
        lw->top_margin = 12;


        /* Create toplevel. */
        if( 
            OSWCreateWindow(
                &lw->toplevel,
                parent,
                lw->x, lw->y, 
                lw->width, lw->height
            )
        )
            return(-1);

        OSWSetWindowInput(
	    lw->toplevel,
	    KeyPressMask | KeyReleaseMask |
            ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
            ExposureMask | VisibilityChangeMask
        );


        /* Scroll bars. */
	if(
            SBarInit(
                &lw->sb,
                lw->toplevel,
                lw->width,
                lw->height
	    )
        )
	    return(-1);


        /* Add this widget to the widget regeristry. */
        WidgetRegAdd(lw, WTYPE_CODE_LIST);


	return(0);
}       



/*
 *	Resizes the list window.
 */
int ListWinResize(list_window_struct *lw)
{
        win_attr_t wattr;


        if(lw == NULL)
            return(-1);


        OSWGetWindowAttributes(lw->toplevel, &wattr);

        /* No change in size? */
        if((lw->width == wattr.width) &&
           (lw->height == wattr.height)
        ) 
            return(0);


        /* Set new size values. */
        lw->x = wattr.x;
        lw->y = wattr.y;
        lw->width = wattr.width;
        lw->height = wattr.height;

        /* Recreate toplevel buffer. */
        OSWDestroyPixmap(&lw->toplevel_buf);
	if(
            OSWCreatePixmap(
                &lw->toplevel_buf,
                lw->width, lw->height
	    )
        )
            return(-1);
            

        /* Resize scroll bars. */
        SBarResize(&lw->sb, lw->width, lw->height);

        SBarDraw(
            &lw->sb,
            lw->width,
            lw->height,
            lw->width,
            (lw->total_entries + 2) * (int)lw->row_height
        ); 


        return(0);
}


/*
 *	Redraws the list window.
 */
int ListWinDraw(list_window_struct *lw)
{
	int i, y_pos;
        int entries_drawn, entries_visible;
        win_attr_t wattr;
	list_window_entry_struct **entry_ptr;
        font_t *prev_font;


        if(lw == NULL)
            return(-1);


        OSWGetWindowAttributes(lw->toplevel, &wattr);

        /* Map as needed. */
        if(!lw->map_state)
        {
            OSWMapWindow(lw->toplevel);
            lw->map_state = 1;

            /* Redraw scroll bars. */
            SBarDraw(
                &lw->sb,
                wattr.width,
                wattr.height,
                wattr.width,
                (lw->total_entries + 2) * (int)lw->row_height
            );
        }

        /* Recreate buffers as needed. */
        if(lw->toplevel_buf == 0)
        {
            if(
                OSWCreatePixmap(
                    &lw->toplevel_buf,
                    wattr.width, wattr.height
                )
            )
                return(-1);
        }


        /* Record previous font. */
        prev_font = OSWQueryCurrentFont();

        OSWSetFont(lw->font);


        /* Clear buffer. */
        if(widget_global.force_mono)
	{
            OSWClearPixmap(
                lw->toplevel_buf,
                wattr.width, wattr.height,
                osw_gui[0].black_pix
            );
	}
        else
	{
            WidgetPutImageTile(
                lw->toplevel_buf,
                widget_global.std_bkg_img,
                wattr.width, wattr.height
            );
	}


        /* Draw entries. */
	i = lw->sb.y_win_pos / (int)lw->row_height;
	if(i >= lw->total_entries)
	    i = (int)lw->total_entries - 1;
	if(i < 0)
	    i = 0;

        y_pos = ((int)lw->top_margin + 9) -
            ((int)lw->sb.y_win_pos % (int)lw->row_height);
        entries_drawn = 0;
	if(lw->row_height > 0)
	    entries_visible = ((int)wattr.height / (int)lw->row_height) + 2;
	else
	    entries_visible = 0;
	entry_ptr = &(lw->entry[i]);

        while((i < lw->total_entries) &&
              (entries_drawn < entries_visible)
	)
        {
            if(*entry_ptr == NULL)
	    {
		i++;
		entry_ptr++;

                continue;
	    }

            /* Skip entries with no name. */
            if((*entry_ptr)->name == NULL)
            {
                i++;
                entry_ptr++;

                y_pos += lw->row_height;
                entries_drawn++;

                continue;
            }
            if((*entry_ptr)->name[0] == '\0')
            {
                i++;
                entry_ptr++;

                y_pos += lw->row_height;
                entries_drawn++;

                continue;
            }

            /* Draw icon image. */
            OSWPutImageToDrawablePos(
                (*entry_ptr)->image,
                (drawable_t)lw->toplevel_buf,
                lw->left_margin, y_pos - 12
            );
            
            /* Is this the selected entry? */
            if(lw->entry_pos == i)
            {
                OSWSetFgPix(widget_global.surface_selected_pix);
                if(((int)lw->row_height - 7) < 1)
                {
                    OSWDrawSolidRectangle(
			lw->toplevel_buf,
                        lw->left_margin + 31, y_pos - 12,
                        strlen((*entry_ptr)->name) * lw->char_width + 6,
                        2
                    );
                }
                else
                {
                    OSWDrawSolidRectangle(
			lw->toplevel_buf,
                        lw->left_margin + 31, y_pos - 12,
                        strlen((*entry_ptr)->name) * lw->char_width + 6,
                        (int)lw->row_height - 7
                    );
                }
                OSWSetFgPix(widget_global.selected_text_pix);
                OSWDrawString(
		    lw->toplevel_buf,
                    lw->left_margin + 33, y_pos + 5,
                    (*entry_ptr)->name
                );
            }
            else
            {
                OSWSetFgPix(widget_global.normal_text_pix);
                OSWDrawString(
		    lw->toplevel_buf,
                    lw->left_margin + 33, y_pos + 5,
                    (*entry_ptr)->name
                );
            }

	    /* Increment index and pointer. */
	    i++;
	    entry_ptr++;

	    /* Increment positions. */
            y_pos += lw->row_height;
            entries_drawn++;
        }

        /* Draw frame depression. */
        if(widget_global.force_mono)
	{
            WidgetFrameButtonPixmap(
                lw->toplevel_buf,
                True,
                wattr.width, wattr.height,
                osw_gui[0].white_pix,
                osw_gui[0].white_pix
            );
	}
	else
	{
            WidgetFrameButtonPixmap(
                lw->toplevel_buf,
                True,
                wattr.width, wattr.height,
                widget_global.surface_highlight_pix,
                widget_global.surface_shadow_pix
            );
	}

        OSWPutBufferToWindow(lw->toplevel, lw->toplevel_buf);

        OSWSetFont(prev_font);


        return(0);
}

/*
 *	Manages list window.
 */
int ListWinManage(
        list_window_struct *lw,
        event_t *event
)
{
	int i, x, y, z, y_max;
        int new_entry_pos;
        list_window_entry_struct *list_entry_ptr;
        win_attr_t wattr;

	int events_handled = 0;

	static long last_click;


        if((lw == NULL) ||
           (event == NULL)
        )
            return(events_handled);


        if(!lw->map_state &&
	   (event->type != MapNotify)
	)
            return(events_handled);


        switch(event->type)
        {
          /* ****************************************************** */
          case ButtonPress:
	    /* Toplevel. */
            if(event->xany.window == lw->toplevel)
            {
                OSWGetWindowAttributes(lw->toplevel, &wattr);

                /* Sanitize lw->row_height. */
                if(lw->row_height < 1) 
                    lw->row_height = 1;

		/* Set in focus. */
                lw->is_in_focus = 1;
		lw->sb.is_in_focus = 1;

                /* Record button state. */
                if(event->xbutton.button == Button1)
                    button1_state = True;
                if(event->xbutton.button == Button2)
                    button2_state = True;
                if(event->xbutton.button == Button3)
                    button3_state = True;

                /* Record button press coordinates. */
                drag_start_x = event->xbutton.x;
                drag_start_y = event->xbutton.y;


		/* Unmap change values prompt. */
		if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
		    ListWinUnmapCVPrompt(lw);


		/* Record last entry position as z. */
		z = lw->entry_pos;

                /* Calculate new entry position. */
                lw->entry_pos = ((int)lw->sb.y_win_pos
                    + (int)event->xbutton.y - (int)lw->top_margin + 6) /
                    (int)lw->row_height;

                /* Sanitize entry position. */
                if(lw->entry_pos >= (int)lw->total_entries)
                    lw->entry_pos = (int)lw->total_entries - 1;
                if(lw->entry_pos < 0)
                    lw->entry_pos = 0;


		/* Scroll as needed if near edge. */
                y_max = (((int)lw->total_entries + 2) *
                    (int)lw->row_height) - wattr.height;
		/* Scroll up? */
		if((event->xbutton.y < (int)lw->row_height) &&
                   (lw->sb.y_win_pos > 0)
		)
		{
		    ListWinDraw(lw);
		    SBarDraw(
                        &lw->sb,
                        wattr.width,
                        wattr.height,
                        wattr.width,
                        (lw->total_entries + 2) * (int)lw->row_height
                    );

		    lw->sb.y_win_pos -= ((int)wattr.height / 2);

                    /* Sanitize scroll. */
                    if(lw->sb.y_win_pos > y_max) 
                        lw->sb.y_win_pos = y_max;
                    if(lw->sb.y_win_pos < 0) 
                        lw->sb.y_win_pos = 0;

		    usleep(300000);

                    SBarDraw(
                        &lw->sb,
                        wattr.width, 
                        wattr.height, 
                        wattr.width,
                        (lw->total_entries + 2) * (int)lw->row_height
                    );
		}
		/* Scroll down? */
		else if((event->xbutton.y > ((int)wattr.height - (int)lw->row_height)) &&
                        (lw->sb.y_win_pos < y_max)
		)
		{
                    ListWinDraw(lw);
                    SBarDraw(
                        &lw->sb,
                        wattr.width,
                        wattr.height,
                        wattr.width, 
                        (lw->total_entries + 2) * (int)lw->row_height
                    );

		    lw->sb.y_win_pos += ((int)wattr.height / 2);

		    /* Sanitize scroll. */
                    if(lw->sb.y_win_pos > y_max)
                        lw->sb.y_win_pos = y_max;
                    if(lw->sb.y_win_pos < 0)
                        lw->sb.y_win_pos = 0;

                    usleep(300000);     

                    SBarDraw(
                        &lw->sb,
                        wattr.width, 
                        wattr.height, 
                        wattr.width,
                        (lw->total_entries + 2) * (int)lw->row_height
                    );
		}

                /* Double clicked? */
                if((lw->entry_pos >= 0) &&
                   (event->xbutton.button == Button1) &&
                   ((last_click + widget_global.double_click_int) >= MilliTime())
                )
                {
                    /* Do select notify procedure. */
                    if((lw->entry_pos == z) &&   
                       (lw->entry_pos >= 0)
                    )
                    {
                        if(lw->func_cb != NULL)
			    lw->func_cb(lw->client_data);
                 
                        /* Reset last_click. */
                        last_click = 0;

                        events_handled++;
                        return(events_handled);
                    }
                    else
                    {
                        /* Record last_click. */ 
                        last_click = MilliTime();   
                    }
                }
                /* Slow double clicked for relabel item? */
                else if((lw->entry_pos >= 0) && 
                        (event->xbutton.button == Button1) &&
                        ((last_click + widget_global.relabel_item_delay) >= MilliTime())
                )
                {
                    /* Do rename procedure. */
                    if((lw->entry_pos == z) &&
                       (lw->entry_pos >= 0)
                    )
                    {
                        ListWinMapCVPrompt(lw, LW_CVP_MODE_RENAME);

                        /* Reset last_click. */
                        last_click = 0;

                        events_handled++;
                        return(events_handled);
                    }
                    else
                    {
                        /* Record last_click. */
                        last_click = MilliTime();
                    }
                }
                else
                {
                    /* Record last click. */
                    last_click = MilliTime();
                }
                events_handled++;
            }
            break;
        
          /* ****************************************************** */
          case ButtonRelease:
            /* Change button states. */
            if(event->xbutton.button == Button1)
                button1_state = False;
            if(event->xbutton.button == Button2)
                button2_state = False;
            if(event->xbutton.button == Button3)
                button3_state = False;


            if(event->xany.window == lw->toplevel)
            {
                /* Manage drag on same window. */
                if(drag_active &&
                   (lw->entry != NULL) &&
                   (lw->toplevel == drag_start_w)
                )
                {
                    /* Sanitize lw->row_height. */
                    if(lw->row_height < 1) 
                        lw->row_height = 1;

                    /* Calculate which position ButtonRelease was on. */
                    new_entry_pos = ((int)lw->sb.y_win_pos
                        + (int)event->xbutton.y - (int)lw->top_margin + 6) /
                        (int)lw->row_height;

                    /* Sanitize entry position. */
                    if(new_entry_pos >= (int)lw->total_entries)
                        new_entry_pos = (int)lw->total_entries - 1;
                    if(new_entry_pos < 0)
                        new_entry_pos = 0;
             
                    if(drag_start_entry_pos >= (int)lw->total_entries)
                        drag_start_entry_pos = (int)lw->total_entries - 1;
                    if(drag_start_entry_pos < 0)
                        drag_start_entry_pos = 0;
        
                    /* Get pointer to drag_start_entry_pos entry. */
                    list_entry_ptr = (list_window_entry_struct *)
                        lw->entry[drag_start_entry_pos];
            
                
                    /* Shift positions. */
                    if(drag_start_entry_pos > new_entry_pos)
                    {
                        for(i = drag_start_entry_pos; i > new_entry_pos; i--)
                        {
                            lw->entry[i] = lw->entry[i - 1];
                        }
                    }
                    else if(drag_start_entry_pos < new_entry_pos)
                    {
                        for(i = drag_start_entry_pos; i < new_entry_pos; i++)
                        {
                            lw->entry[i] = lw->entry[i + 1];
                        }
                    }
                    else
                    {
                        /* Nothing to do if positions are the same. */
                    }
                        
                    /* Set new entry pointer. */
                    lw->entry[new_entry_pos] = (list_window_entry_struct *)
                        list_entry_ptr;
   
                    /* Change entry position. */
                    lw->entry_pos = new_entry_pos;
                }
                    
                events_handled++;
            }

	    /* Set drag to inactive on ButtonPress. */
	    if(drag_active)
	    {                         
                /* Set drag as inactive after a ButtonRelease. */
                OSWUngrabPointer();
	        ListWinRepeatRecordClear();

                drag_active = False;
	        drag_start_w = 0;
            }
            break;

          /* ****************************************************** */
          case MotionNotify:
            if(event->xany.window == lw->toplevel)
            {
                /* Was not previously in drag? */
                if(!drag_active &&
                   button1_state &&
                   lw->allow_drag	/* This list allows drag? */
                )
                {
                    /* Check drag tolorance. */
                    x = event->xmotion.x;
                    y = event->xmotion.y;
                    if((x < (drag_start_x - (LW_DRAG_TOLORANCE / 2))) ||
                       (x >= (drag_start_x + (LW_DRAG_TOLORANCE / 2))) ||
                       (y < (drag_start_y - (LW_DRAG_TOLORANCE / 2))) ||
                       (y >= (drag_start_y + (LW_DRAG_TOLORANCE / 2)))
                    )
                    {
                        /* Drag starts, set up drag values. */
                        drag_start_entry_pos = lw->entry_pos;
                        drag_start_w = event->xany.window;
                        drag_active = True;

			/* Grab pointer. */
			OSWGrabPointer(
			    lw->toplevel,
			    True,
			    ButtonReleaseMask | PointerMotionMask,
                            GrabModeAsync, GrabModeAsync,
                            None,
			    (widget_global.drag_item_wcr == NULL) ?
				0 :
                                widget_global.drag_item_wcr->cursor
			);

                        ListWinRepeatRecordSet(
                            lw,
/* Use popup list delay for now. */
                            widget_global.pulist_repeat_delay,
                            LW_OP_CODE_NONE
                        );
                    }
                }
            }                 
            break;

          /* ****************************************************** */
          case KeyPress:
            /* Skip if not in focus. */
            if(!lw->is_in_focus)
                break;

	    /* Escape (for cv_prompt). */
            if(event->xkey.keycode == osw_keycode.esc)
	    {
		if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
		{
		    ListWinUnmapCVPrompt(lw);

		    events_handled++;
		    return(events_handled);
		}
	    }
            /* Enter (for cv_prompt). */
            else if((event->xkey.keycode == osw_keycode.enter) ||
                    (event->xkey.keycode == osw_keycode.np_enter)
	    )
            {
                if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
                {   
                    ListWinCVPromptApply(lw);

                    ListWinUnmapCVPrompt(lw);

		    ListWinDraw(lw);

                    events_handled++;
                    return(events_handled);
                }
            }

	    /* Do not handle any more types of keys if in cv mode. */
	    if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
		break;

            /* Up. */
            if(event->xkey.keycode == osw_keycode.cursor_up)
            {
                /* Sanitize lw->row_height. */
                if(lw->row_height < 1)
                    lw->row_height = 1;

                /* Move selected entry position up. */
                lw->entry_pos--;

                /* Sanitize entry position. */  
                if(lw->entry_pos >= (int)lw->total_entries)
                    lw->entry_pos = (int)lw->total_entries - 1;
                if(lw->entry_pos < 0)
                    lw->entry_pos = 0;

                /* Move scrolled position as needed. */
                if((lw->sb.y_win_pos / (int)lw->row_height) > lw->entry_pos)
                {
                    lw->sb.y_win_pos = lw->entry_pos * (int)lw->row_height;

                    /* Redraw scroll bars. */
                    SBarDraw(
                        &lw->sb,
                        wattr.width,
                        wattr.height,
                        wattr.width,
                        (lw->total_entries + 2) * (int)lw->row_height
                    );
                }

                events_handled++;
            }
            /* Down. */
            else if(event->xkey.keycode == osw_keycode.cursor_down)
            {
                /* Sanitize lw->row_height. */
                if(lw->row_height < 1)
                    lw->row_height = 1;

                /* Move selected entry position down. */
                lw->entry_pos++;

                /* Sanitize entry position. */
                if(lw->entry_pos >= (int)lw->total_entries)
                    lw->entry_pos = (int)lw->total_entries - 1;
                if(lw->entry_pos < 0)
                    lw->entry_pos = 0;

                /* Move scrolled position as needed. */
                OSWGetWindowAttributes(lw->toplevel, &wattr);
                if(((int)lw->sb.y_win_pos + (int)wattr.height) <=
                   ((lw->entry_pos + 1) * (int)lw->row_height)
                )
                {
                    lw->sb.y_win_pos = ((lw->entry_pos + 1) *
                        (int)lw->row_height) - (int)wattr.height + 8;

                    if((int)lw->sb.y_win_pos < 0)
                        lw->sb.y_win_pos = 0;

                    /* Redraw scroll bars. */
                    SBarDraw(
                        &lw->sb,
                        wattr.width,
                        wattr.height,
                        wattr.width,
                        (lw->total_entries + 2) * (int)lw->row_height
                    );
                }

                events_handled++;
	    }
            /* Enter. */
            else if((event->xkey.keycode == osw_keycode.enter) ||
                    (event->xkey.keycode == osw_keycode.np_enter)
            )
            {
                if(lw->func_cb != NULL)
                {
		    lw->func_cb(lw->client_data);
                    
                    events_handled++;
                    return(events_handled);
                }
            }
            break;

          /* ****************************************************** */
          case KeyRelease:
	    if(!lw->is_in_focus)
		break;

	    break;

          /* ****************************************************** */
          case Expose:
            if(event->xany.window == lw->toplevel)
            {
                events_handled++;
            }
            break;

          /* ****************************************************** */
          case FocusIn:
            if(event->xany.window == lw->toplevel)
            {
                lw->is_in_focus = 1;

                events_handled++;
		return(events_handled);
            }
            break;

          /* ****************************************************** */
          case FocusOut:
            if(event->xany.window == lw->toplevel)
            {
                lw->is_in_focus = 0;

                events_handled++;
		return(events_handled);
            }
            break;

          /* ****************************************************** */
          case VisibilityNotify:
            if(event->xany.window == lw->toplevel)
            {
                lw->visibility_state = event->xvisibility.state;

                events_handled++;
                return(events_handled);
            }
            break;
        }


        /* ******************************************************** */

        /* Scroll bar. */
        if(events_handled == 0)
        {
            events_handled += SBarManage(
                &lw->sb,
                lw->width,
                lw->height,
                lw->width,
                (lw->total_entries + 2) * (int)lw->row_height,
                event
            );
        }
        /* Redraw as needed. */
        if(events_handled > 0)
        {
            ListWinDraw(lw);
        }

        /* Change values prompt. */
        if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
        {
            events_handled += PromptManage(&lw->cv_prompt, event);
        }


        return(events_handled);
}



/*
 *	Applies changes on the cv prompt.
 */
int ListWinCVPromptApply(list_window_struct *lw)
{
	int i, len;


	if(lw == NULL)
	    return(-1);


	/* Not in cv prompt mode? */
	if(lw->cv_prompt_mode == LW_CVP_MODE_NONE)
	    return(-1);
	if(lw->cv_prompt.buf == NULL)
	    return(-1);

	i = lw->cv_prompt_entry_pos;
	if(!ListIsItemAllocated(lw, i))
	    return(-1);

	len = strlen(lw->cv_prompt.buf);
	free(lw->entry[i]->name);
	lw->entry[i]->name = (char *)calloc(1, (len + 1) * sizeof(char));
	if(lw->entry[i]->name == NULL)
	    return(-1);

	strcpy(lw->entry[i]->name, lw->cv_prompt.buf);



	return(0);
}

/*
 *	Map change values prompt on list.
 */
void ListWinMapCVPrompt(list_window_struct *lw, int mode)
{
        int i, x, y;
        unsigned int width, height;


        if(lw == NULL)
            return;


        /* Already mapped? */
        if(lw->cv_prompt_mode != LW_CVP_MODE_NONE)
            return;

	/* Is selected entry valid? */
	if(!ListIsItemAllocated(lw, lw->entry_pos))
	    return;


        /* Initialize change value prompt. */
        if(
            PromptInit(
                &lw->cv_prompt,
                lw->toplevel,  
                10,
                ((int)lw->height / 2) - (LW_PROMPT_HEIGHT / 2),
                MAX((int)lw->width - 20, 100),
                LW_PROMPT_HEIGHT,
                PROMPT_STYLE_FLUSHED,
                NULL,
                1024,			/* Buffer length. */
                0,			/* Number of history buffers. */
                NULL
            )
        )
            return;


        if((lw->map_state == 0) ||
           (lw->cv_prompt.buf == NULL)
        )
            return;


        /* Change focus on lw and its widgets. */
        lw->is_in_focus = 1;

        lw->sb.is_in_focus = 0;


        /* ********************************************************** */

	/* Record selected entry. */
	lw->cv_prompt_entry_pos = lw->entry_pos;

	switch(mode)
	{
          case LW_CVP_MODE_RENAME:
	    /* Copy entry's name to cv prompt buffer. */
	    i = lw->cv_prompt_entry_pos;
            if(lw->entry[i]->name != NULL)
	    {
		strncpy(
		    lw->cv_prompt.buf,
		    lw->entry[i]->name,
		    lw->cv_prompt.buf_len
		);


		x = lw->left_margin + 28;
		y = (lw->row_height * i) - lw->sb.y_win_pos
		    + lw->top_margin - 8;

		width = 200;
		height = LW_PROMPT_HEIGHT;

                /*   Move the change values prompt to the right
                 *   position and map it.
                 */
                lw->cv_prompt_mode = mode;
                OSWReparentWindow(lw->cv_prompt.toplevel, lw->toplevel);
                OSWMoveResizeWindow(lw->cv_prompt.toplevel, x, y, width, height);
		PromptUnmarkBuffer(&lw->cv_prompt, PROMPT_POS_END);
                PromptMap(&lw->cv_prompt);

                lw->cv_prompt.is_in_focus = 1;
	    }
	    break;


	}


	return;
}

/*
 *	Unmap change values prompt on list window.
 */
void ListWinUnmapCVPrompt(list_window_struct *lw)          
{
        if(lw == NULL)
            return;   


        /* Reset target item. */
        lw->cv_prompt_entry_pos = -1;

        /* Reset mode. */
        lw->cv_prompt_mode = LW_CVP_MODE_NONE;

        /* Destroy change values prompt. */
        PromptDestroy(&lw->cv_prompt);
                 
        /* Make sure it is out of focus. */
        lw->cv_prompt.is_in_focus = 0;


        return;
}


/*
 *	List window map.
 */
void ListWinMap(list_window_struct *lw)
{
        if(lw == NULL)
            return;


        lw->map_state = 0;
        ListWinDraw(lw);

        /* Reset drag globals. */
        button1_state = False;
        button2_state = False;
        button3_state = False;
        drag_active = False;


        return;
}

/*
 *	List window unmap.
 */
void ListWinUnmap(list_window_struct *lw)
{
	if(lw == NULL)
            return;


        OSWUnmapWindow(lw->toplevel);
        lw->map_state = 0;
        lw->visibility_state = VisibilityFullyObscured;
        lw->is_in_focus = 0;


        /* Destroy buffers. */
        OSWDestroyPixmap(&lw->toplevel_buf);


        return;    
}


/*
 *	List window destroy.
 */
void ListWinDestroy(list_window_struct *lw)
{  
        if(lw == NULL)
	    return;


        /* Delete widget from regeristry. */
        WidgetRegDelete(lw);


        /* Free all list items. */
        ListDeleteAllItems(lw);


        if(IDC())  
        {
            PromptDestroy(&lw->cv_prompt);

            SBarDestroy(&lw->sb);
            
            OSWDestroyPixmap(&lw->toplevel_buf);
            OSWDestroyWindow(&lw->toplevel);
        }


        lw->map_state = 0;
        lw->visibility_state = VisibilityFullyObscured;
        lw->is_in_focus = 0;  
        lw->x = 0;
        lw->y = 0;
        lw->width = 0;
        lw->height = 0;
	lw->font = NULL;
        lw->next = NULL;
        lw->prev = NULL;


        /* Clear repeat record as needed. */
        if(lw == lw_repeat_record[0].lw)
            ListWinRepeatRecordClear();


        return;
}



/*
 *	Sets the list window repeat records.
 */
void ListWinRepeatRecordSet(
	list_window_struct *lw,
	long dsec, int op_code
)
{
        lw_repeat_record[0].lw = lw;
        lw_repeat_record[0].next_repeat = MilliTime() + MAX(dsec, 0);
        lw_repeat_record[0].op_code = op_code;

	return;
}


/*
 *	Clears the list window repeat record.
 */
void ListWinRepeatRecordClear()
{
	lw_repeat_record[0].lw = NULL;
        lw_repeat_record[0].next_repeat = 0;
        lw_repeat_record[0].op_code = LW_OP_CODE_NONE;

        return;
}


/*
 *	Manages list window repeats.
 */
int ListWinManageRepeat(event_t *event) 
{
        int x, y, root_x, root_y, y_max;
        int events_handled = 0;
        list_window_struct *lw;
        win_attr_t wattr;

 
        /*
         *   Note: The event information is not used and not needed here. 
         */

        /* Is a list pointer set in the record? */
        if(lw_repeat_record[0].lw != NULL)  
        {
            lw = lw_repeat_record[0].lw;

            /* Time due for repeat? */
            if(MilliTime() >= lw_repeat_record[0].next_repeat)
            {
                /* Check pointer's position. */
                OSWGetWindowAttributes(lw->toplevel, &wattr);

                OSWGetPointerCoords(
                    lw->toplevel,
                    &root_x, &root_y,
                    &x, &y
                );
                if(y < 0)
                {
                    lw_repeat_record[0].op_code = 
                        LW_OP_CODE_SCROLL_UP;
                }
                else if(y >= (int)wattr.height)
                {
                    lw_repeat_record[0].op_code =
                        LW_OP_CODE_SCROLL_DOWN;
                }
                else
                {
                    lw_repeat_record[0].op_code = LW_OP_CODE_NONE;
                }

                /* Handle scroll operation. */
                switch(lw_repeat_record[0].op_code)
                {
                  case LW_OP_CODE_SCROLL_UP:  
                    lw->sb.y_win_pos -= lw->row_height;
                    events_handled++;
                    break;

                  case LW_OP_CODE_SCROLL_DOWN:
                    lw->sb.y_win_pos += lw->row_height;
                    events_handled++;
                    break;

                  case LW_OP_CODE_NONE:
/* do not clear, only ButtonRelease should do that.
                    ListWinRepeatRecordClear();
*/
                    return(events_handled);
                    break;

                  default:
                    break;
                }

                /* Sanitize position. */
                y_max = (((int)lw->total_entries + 2) *
		    (int)lw->row_height) - wattr.height;
                if(lw->sb.y_win_pos > y_max)
                    lw->sb.y_win_pos = y_max;
                if(lw->sb.y_win_pos < 0)
                    lw->sb.y_win_pos = 0;

                /* Redraw. */
                ListWinDraw(lw);

                /* Redraw scroll bars. */
                SBarDraw(
                    &lw->sb,
                    wattr.width,
                    wattr.height,
                    wattr.width, 
                    (lw->total_entries + 2) * (int)lw->row_height
                );

                    
                /* Schedual next operation. */
                lw_repeat_record[0].next_repeat =
/* Use pulist repeat for now. */
                    MilliTime() + widget_global.pulist_repeat_interval;


                events_handled++;
            }
        }


        return(events_handled);
}
